import json
import os
import pathlib
from itertools import groupby
from operator import attrgetter
from typing import TYPE_CHECKING

import numpy as np

from ada import Beam, Material, Part, Section
from ada.concepts.containers import Nodes
from ada.config import get_logger

from .utils import to_safe_name

if TYPE_CHECKING:
    from ada import FEM, Assembly

logger = get_logger()


def write_assembly_to_cache(assembly: "Assembly", cache_file_path):
    """Write the Assembly information to a HDF5 file format for High performance cache."""
    import h5py

    # TODO: Add support for FEM only to begin with

    cache_file_path = pathlib.Path(cache_file_path)
    os.makedirs(cache_file_path.parent, exist_ok=True)
    h5_filename = cache_file_path.with_suffix(".h5")
    with h5py.File(h5_filename, "w") as f:
        info = f.create_group("INFO")
        info.attrs.create("NAME", assembly.name)

        parts_group = f.create_group("PARTS")

        walk_parts(parts_group, assembly)

    print(f'Saved cached model at "{cache_file_path}"')


def walk_parts(cache_p, part):
    for p in part.parts.values():
        part_group = add_part_to_cache(p, cache_p)
        part_group.attrs.create("PARENT", to_safe_name(p.parent.name))
        walk_parts(part_group, p)


def add_part_to_cache(part: Part, parent_part_group):
    part_group = parent_part_group.create_group(to_safe_name(part.name))

    part_group.attrs.create("METADATA", json.dumps(part.metadata))

    if len(part.nodes) > 0:
        add_nodes_to_cache(part.nodes, part_group)

    if len(part.sections) > 0:
        add_sections_to_cache(part, part_group)

    if len(part.materials) > 0:
        add_materials_to_cache(part, part_group)

    if len(part.beams) > 0:
        add_beams_to_cache(part, part_group)

    if len(part.plates) > 0:
        add_plates_to_cache()

    if len(part.shapes) > 0:
        add_shapes_to_cache()

    if len(part.pipes) > 0:
        add_pipes_to_cache()

    if len(part.walls) > 0:
        add_walls_to_cache()

    # Add FEM object
    if len(part.fem.nodes) > 0:
        print(f'Caching FEM data from "{part.name}"')
        add_fem_to_cache(part.fem, part_group)

    return part_group


def add_sections_to_cache(part, parts_group):
    prefix = "SECTIONS"

    def add_ints_to_cache(s: Section):
        return [x if x is not None else 0 for x in [s.r, s.wt, s.h, s.w_top, s.w_btn, s.t_w, s.t_ftop, s.t_fbtn, s.id]]

    def add_strings_to_cache(s: Section):
        return [s.guid, s.name, s.units.value, s.type.value]

    parts_group.create_dataset(f"{prefix}_STR", data=[add_strings_to_cache(bm) for bm in part.sections])
    parts_group.create_dataset(f"{prefix}_INT", data=[add_ints_to_cache(bm) for bm in part.sections])


def add_materials_to_cache(part, parts_group):
    prefix = "MATERIALS"

    def add_ints_to_cache(e: Material):
        m = e.model
        return [m.E, m.rho, m.sig_y, e.id]

    def add_strings_to_cache(e: Material):
        return [e.guid, e.name, e.units.value]

    parts_group.create_dataset(f"{prefix}_INT", data=[add_ints_to_cache(bm) for bm in part.materials])
    parts_group.create_dataset(f"{prefix}_STR", data=[add_strings_to_cache(bm) for bm in part.materials])


def add_plates_to_cache():
    logger.error("Plate caching is not yet implemented")


def add_shapes_to_cache():
    logger.error("Shape caching is not yet implemented")


def add_pipes_to_cache():
    logger.error("Pipes caching is not yet implemented")


def add_walls_to_cache():
    logger.error("Walls caching is not yet implemented")


def add_beams_to_cache(part: Part, parts_group):
    prefix = "BEAMS"

    def add_int_cache(bm: Beam, up=False):
        if up is True:
            nids = bm.up.tolist()
        else:
            nids = [bm.n1.id, bm.n2.id]
        if None in nids:
            raise ValueError()
        return nids

    def add_str_cache(bm: Beam):
        return [bm.guid, bm.name, bm.section.name, bm.material.name, json.dumps(bm.metadata)]

    parts_group.create_dataset(f"{prefix}_INT", data=[add_int_cache(bm) for bm in part.beams])
    parts_group.create_dataset(f"{prefix}_STR", data=[add_str_cache(bm) for bm in part.beams])
    parts_group.create_dataset(f"{prefix}_UP", data=[add_int_cache(bm, True) for bm in part.beams])


def add_fem_to_cache(fem: "FEM", part_group):
    fem_group = part_group.create_group("FEM")
    fem_group.attrs.create("NAME", to_safe_name(fem.name))

    # Add Nodes
    add_nodes_to_cache(fem.nodes, fem_group)

    # Add elements
    elements_group = fem_group.create_group("MESH")
    for group, elements in groupby(sorted(fem.elements, key=attrgetter("type")), key=attrgetter("type")):
        med_cells = elements_group.create_group(group.value)
        med_cells.create_dataset("ELEMENTS", data=[[int(el.id), *[int(n.id) for n in el.nodes]] for el in elements])


def add_nodes_to_cache(nodes: Nodes, group):
    points = np.array([[n.id, *n.p] for n in nodes])
    coo = group.create_dataset("NODES", data=points)
    coo.attrs.create("NBR", len(points))



from ada.fem.shapes import ElemShape

sh = ElemShape.TYPES.shell
so = ElemShape.TYPES.solids
li = ElemShape.TYPES.lines

ada_to_meshio = {
    sh.TRI: "triangle",
    sh.TRI6: "triangle6",
    sh.TRI7: "triangle7",
    sh.QUAD: "quad",
    sh.QUAD8: "quad8",
    so.HEX8: "hexahedron",
    so.HEX20: "hexahedron20",
    so.HEX27: "hexahedron27",
    so.TETRA: "tetra",
    so.TETRA10: "tetra10",
    so.PYRAMID5: "pyramid5",
    so.WEDGE: "wedge",
    so.WEDGE15: "wedge15",
    li.LINE: "line",
    li.LINE3: "line3",
}

meshio_to_ada = {val: key for key, val in ada_to_meshio.items()}



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, List, Union

import numpy as np

from ada.base.units import Units
from ada.config import Settings, get_logger
from ada.core.vector_utils import vector_length

if TYPE_CHECKING:
    from ada import Beam
    from ada.concepts.containers import Nodes
    from ada.fem import Bc, Csys, Elem

numeric = Union[int, float, np.number]
logger = get_logger()


class Node:
    """Base node object"""

    def __init__(
        self, p: Iterable[numeric, numeric, numeric], nid=None, bc=None, r=None, parent=None, units=Units.M, refs=None
    ):
        self._id = nid
        self.p: np.ndarray = np.array([*p], dtype=np.float64) if type(p) != np.ndarray else p
        if len(self.p) != 3:
            raise ValueError("Node object must have exactly 3 coordinates (x, y, z).")

        self._bc = bc
        self._r = r
        self._parent = parent
        self._units = units
        self._refs = [] if refs is None else refs

    @property
    def id(self) -> int:
        return self._id

    @id.setter
    def id(self, value: int):
        self._id = value

    @property
    def x(self):
        return self.p[0]

    @property
    def y(self):
        return self.p[1]

    @property
    def z(self):
        return self.p[2]

    @property
    def bc(self) -> Bc:
        return self._bc

    @bc.setter
    def bc(self, value: Bc):
        self._bc = value

    @property
    def r(self) -> float:
        return self._r

    def p_roundoff(self, scale_factor: Union[int, float] = 1, precision: int = Settings.precision) -> None:
        from ada.core.utils import roundoff

        self.p = np.array([roundoff(scale_factor * x, precision=precision) for x in self.p])

    def add_obj_to_refs(self, item) -> None:
        if item not in self.refs:
            self.refs.append(item)
        else:
            logger.debug(f"Item {item} is already in node refs {self}")

    def remove_obj_from_refs(self, item) -> None:
        if item in self.refs:
            self.refs.remove(item)
        else:
            logger.debug(f"Item {item} is not in node refs {self}")

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            scale_factor = Units.get_scale_factor(self._units, value)

            self.p_roundoff(scale_factor)

            if self._r is not None:
                self._r *= scale_factor
            self._units = value

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def refs(self) -> List[Union[Elem, Beam, Csys]]:
        return self._refs

    @property
    def has_refs(self) -> bool:
        """Returns if node is valid, i.e. has objects in refs"""
        return len(self.refs) > 0

    def get_main_node_at_point(self) -> Node:
        nodes = self.sort_by_refs_at_point()
        (nearest_node,) = sort_nodes_by_distance(self, nodes)
        return nearest_node

    def sort_by_refs_at_point(self) -> list[Node]:
        nodes = list(filter(lambda n: n.has_refs, self.parent.nodes.get_by_volume(self)))
        if len(nodes) > 0:
            return sorted(nodes, key=lambda n: len(n.refs), reverse=True)
        else:
            return [self]

    def __getitem__(self, index):
        return self.p[index]

    def __gt__(self, other):
        return tuple(self.p) > tuple(other.p)

    def __lt__(self, other):
        return tuple(self.p) < tuple(other.p)

    def __ge__(self, other):
        return tuple(self.p) >= tuple(other.p)

    def __le__(self, other):
        return tuple(self.p) <= tuple(other.p)

    def __eq__(self, other: Node):
        if not isinstance(other, Node):
            return NotImplemented
        return (*self.p, self.id) == (*other.p, other.id)

    def __ne__(self, other: Node):
        if not isinstance(other, Node):
            return NotImplemented
        return (*self.p, self.id) != (*other.p, other.id)

    def __hash__(self):
        return hash((*self.p, self.id))

    def __repr__(self):
        return f"Node([{self.x}, {self.y}, {self.z}], {self.id})"


def get_singular_node_by_volume(nodes: Nodes, p: np.ndarray, tol=Settings.point_tol) -> Node:
    """Returns existing node within the volume, or creates and returns a new Node at the point"""
    nds = nodes.get_by_volume(p, tol=tol)
    if len(nds) > 0:
        node, *other_nodes = nds
        if len(other_nodes) > 0:
            logger.warning(f"More than 1 node within point {p}, other nodes: {other_nodes}. Returns node {node}")
        return node
    else:
        return Node(p)


def sort_nodes_by_distance(point: Union[Node, np.ndarray], nodes: list[Node]) -> list[Node]:
    if isinstance(point, Node):
        point = point.p
    return sorted(nodes, key=lambda x: vector_length(x.p - point))


def replace_nodes_by_tol(nodes, decimals=0, tol=Settings.point_tol):
    """

    :param nodes:
    :param decimals:
    :param tol:
    :type nodes: ada.core.containers.Nodes
    """

    def rounding(vec, decimals_):
        return np.around(vec, decimals=decimals_)

    def n_is_most_precise(n, nearby_nodes_, decimals_=0):
        most_precise = [np.array_equal(n.p, rounding(n.p, decimals_)) for n in [node] + nearby_nodes_]

        if most_precise[0] and not np.all(most_precise[1:]):
            return True
        elif not most_precise[0] and np.any(most_precise[1:]):
            return False
        elif decimals_ == 10:
            logger.error(f"Recursion started at 0 decimals, but are now at {decimals_} decimals. Will proceed with n.")
            return True
        else:
            return n_is_most_precise(n, nearby_nodes_, decimals_ + 1)

    for node in nodes:
        nearby_nodes = list(filter(lambda x: x != node, nodes.get_by_volume(node.p, tol=tol)))
        if nearby_nodes and n_is_most_precise(node, nearby_nodes, decimals):
            for nearby_node in nearby_nodes:
                replace_node(nearby_node, node)


def replace_node(old_node: Node, new_node: Node) -> None:
    """
    Exchange the old nod with the new. The refs in old node is cleared, and added to new node ref
    :param old_node:
    :param new_node:
    """

    for obj in old_node.refs.copy():
        obj: Union[Beam, Csys, Elem]
        obj.updating_nodes(old_node, new_node)

        old_node.remove_obj_from_refs(obj)
        new_node.add_obj_to_refs(obj)

        logger.debug(f"{old_node} exchanged with {new_node} --> {obj}")



from __future__ import annotations

from bisect import bisect_left
from dataclasses import dataclass
from functools import partial
from itertools import chain, groupby
from operator import attrgetter
from typing import TYPE_CHECKING, Dict, Iterable, List, Tuple, Union

import numpy as np

from ada.concepts.containers import Materials
from ada.concepts.points import Node
from ada.config import get_logger
from ada.core.utils import Counter
from ada.fem.elements import Connector, Elem, Mass, MassTypes
from ada.fem.exceptions.model_definition import FemSetNameExists
from ada.fem.sections import FemSection
from ada.fem.sets import FemSet, SetTypes
from ada.fem.shapes import ElemType
from ada.materials import Material
from ada.sections import Section

if TYPE_CHECKING:
    from ada import FEM
    from ada.fem.results.common import ElementBlock

logger = get_logger()


@dataclass
class COG:
    p: np.array
    tot_mass: float = None
    tot_vol: float = None
    sh_mass: float = None
    bm_mass: float = None
    no_mass: float = None

    @property
    def x(self) -> float:
        """Returns x-coordinate to the point p"""
        return self.p[0]

    @property
    def y(self) -> float:
        """Returns y-coordinate to the point p"""
        return self.p[1]

    @property
    def z(self) -> float:
        """Returns z-coordinate to the point p"""
        return self.p[2]


class FemElements:
    """Container class for FEM elements"""

    def __init__(
        self, elements: Iterable[Union[Elem, Mass, Connector]] = None, fem_obj: FEM = None, from_np_array=None
    ):
        self._fem_obj = fem_obj
        if from_np_array is not None:
            elements = self.elements_from_array(from_np_array)

        self._elements = list(sorted(elements, key=attrgetter("id"))) if elements is not None else []
        self._idmap = {e.id: e for e in self._elements} if len(self._elements) > 0 else dict()

        if len(self._idmap) != len(self._elements):
            raise ValueError("Unequal length of idmap and elements. Might indicate doubly defined element id's")

        self._by_types = None
        self._group_by_types()

    def renumber(self, start_id=1, renumber_map: dict = None):
        """Ensures that the node numberings starts at 1 and has no holes in its numbering."""
        if renumber_map is not None:
            self._renumber_from_map(renumber_map)
        else:
            self._renumber_linearly(start_id)

        self._idmap = {e.id: e for e in self._elements} if len(self._elements) > 0 else dict()
        self._group_by_types()

    def _renumber_from_map(self, renumber_map):
        for el in sorted(self._elements, key=attrgetter("id")):
            if isinstance(el, Mass) or el.type == Elem.EL_TYPES.MASS_SHAPES.MASS:
                # Mass elements are points and have been renumbered during node-renumbering
                continue
            el.id = renumber_map[el.id]

    def _renumber_linearly(self, start_id):
        elid = Counter(start_id)
        for el in sorted(self._elements, key=attrgetter("id")):
            el.id = next(elid)

    def link_nodes(self):
        """Link element nodes with the parent fem node collection"""

        def grab_nodes(elem: Elem):
            nodes = [self._fem_obj.nodes.from_id(no) for no in elem.nodes if type(no) in (int, np.int32)]
            if len(nodes) != len(elem.nodes):
                raise ValueError("Unable to convert element nodes")
            elem._nodes = nodes

        list(map(grab_nodes, self._elements))

    def _build_sets_from_elsets(self, elset, elem_iter):
        if elset is not None and type(elset) == str:

            def link_elset(elem, elem_set_):
                elem._elset = elem_set_

            elements = [self.from_id(el.id) for el in elem_iter]
            if elset not in self._fem_obj.elsets:
                elem_set = FemSet(elset, elements, "elset", parent=self._fem_obj)
                self._fem_obj.sets.add(elem_set)
                list(map(lambda x: link_elset(x, elem_set), elements))
            else:
                self._fem_obj.elsets[elset].add_members(elements)
                elem_set = self._fem_obj.elsets[elset]
                list(map(lambda x: link_elset(x, elem_set), elements))

    def elements_from_array(self, array):
        def to_elem(e):
            nodes = [self._fem_obj.nodes.from_id(n) for n in e[3:] if (n == 0 or np.isnan(n)) is False]
            return Elem(e[0], nodes, e[1], e[2], parent=self._fem_obj)

        return list(map(to_elem, array))

    def build_sets(self):
        """Create sets from attached elset attribute on elements"""
        for elset, elements in groupby(self._elements, key=attrgetter("elset")):
            if elset is None:
                continue
            self._build_sets_from_elsets(elset, elements)

    def remove_elements_by_set(self, elset: FemSet):
        """Remove elements from element set. Will remove element set on completion"""
        for el in elset.members:
            self.remove(el)
        self._sort()
        p = elset.parent
        p.sets.elements.pop(elset.name)

    def remove_elements_by_id(self, ids: Union[int, List[int]]):
        """Remove elements from element ids. Will remove elements on completion"""
        ids = list(ids) if isinstance(ids, Iterable) else [ids]
        for elem_id in ids:
            self.remove(self._idmap[elem_id])
        self._sort()

    def to_elem_blocks(self) -> list[ElementBlock]:
        from ada.fem.results.common import ElementBlock, ElementInfo, FEATypes

        elements = []
        for el_type, el_group in self.group_by_type():
            info = ElementInfo(el_type, FEATypes.GMSH, None)
            elem_data = np.array([tuple([e.id, *[n.id for n in e.nodes]]) for e in el_group], dtype=int)
            el_identifiers = elem_data[:, 0]
            node_refs = elem_data[:, 1:]
            block = ElementBlock(info, node_refs, el_identifiers)
            elements.append(block)
        return elements

    def __contains__(self, item: Elem):
        return item in self._elements

    def __len__(self):
        return len(self._elements)

    def __iter__(self):
        return iter(self._elements)

    def __getitem__(self, index):
        result = self._elements[index]
        return FemElements(result) if isinstance(index, slice) else result

    def __add__(self, other: FemElements):
        other.renumber(self.max_el_id + 1)
        for el in other.elements:
            el.parent = self.parent

        return FemElements(chain.from_iterable([self.elements, other.elements]), self.parent)

    def __repr__(self):
        data = {}
        for key, val in groupby(sorted(self._elements, key=attrgetter("type")), key=attrgetter("type")):
            if key not in data.keys():
                data[key] = len(list(val))
            else:
                data[key] += len(list(val))
        data_str = ", ".join([f'"{key}": {val}' for key, val in data.items()])
        return f"FemElementsCollection(Elements: {len(self._elements)}, By Type: {data_str})"

    def by_types(self):
        return self._by_types

    def calc_cog(self) -> COG:
        """Calculate COG of your FEM model based on element mass distributed to element and nodes"""
        from itertools import chain

        from ada.core.vector_utils import global_2_local_nodes, poly_area, vector_length

        def calc_sh_elem(el: Elem):
            locx = el.fem_sec.local_x
            locy = el.fem_sec.local_y
            locz = el.fem_sec.local_z
            origin = el.nodes[0]
            t = el.fem_sec.thickness

            ln = global_2_local_nodes([locx, locy, locz], origin, el.nodes)

            x, y, z = list(zip(*ln))
            area = poly_area(x, y)
            vol_ = t * area
            mass = vol_ * el.fem_sec.material.model.rho
            center = sum([e.p for e in el.nodes]) / len(el.nodes)

            return mass, center, vol_

        def calc_bm_elem(el: Elem):
            nodes_ = el.get_offset_coords()
            elem_len = vector_length(nodes_[-1] - nodes_[0])
            vol_ = el.fem_sec.section.properties.Ax * elem_len
            mass = vol_ * el.fem_sec.material.model.rho
            center = sum([e.p for e in el.nodes]) / len(el.nodes)

            return mass, center, vol_

        def calc_mass_elem(el: Mass):
            if el.type != MassTypes.MASS:
                raise NotImplementedError(f'Mass type "{el.mass_props.type}" is not yet implemented')
            mass = el.mass
            vol_ = 0.0
            return mass, el.nodes[0].p, vol_

        sh = list(chain(map(calc_sh_elem, self.shell)))
        bm = list(chain(map(calc_bm_elem, self.lines)))
        ma = list(chain(map(calc_mass_elem, self.masses)))

        tot_mass = 0.0
        tot_vol = 0.0
        mcog_ = np.array([0, 0, 0]).astype(float)

        sh_mass = sum([r[0] for r in sh])
        bm_mass = sum([r[0] for r in bm])
        no_mass = sum([r[0] for r in ma])

        for m, c, vol in sh + bm + ma:
            tot_vol += vol
            tot_mass += m
            mcog_ += m * np.array(c).astype(float)

        cog_ = mcog_ / tot_mass

        return COG(cog_, tot_mass, tot_vol, sh_mass, bm_mass, no_mass)

    @property
    def parent(self) -> FEM:
        return self._fem_obj

    @parent.setter
    def parent(self, value: FEM):
        self._fem_obj = value

    @property
    def max_el_id(self):
        if len(self._idmap.keys()) == 0:
            return 0

        return max(self._idmap.keys())

    @property
    def min_el_id(self):
        if len(self._idmap.keys()) == 0:
            return 0

        return min(self._idmap.keys())

    @property
    def elements(self) -> List[Union[Elem, Connector, Mass]]:
        return self._elements

    @property
    def solids(self) -> Iterable[Elem]:
        return filter(lambda x: isinstance(x.type, Elem.EL_TYPES.SOLID_SHAPES), self.stru_elements)

    @property
    def shell(self) -> Iterable[Elem]:
        return filter(lambda x: isinstance(x.type, Elem.EL_TYPES.SHELL_SHAPES), self.stru_elements)

    @property
    def lines(self) -> Iterable[Elem]:
        return filter(lambda x: isinstance(x.type, Elem.EL_TYPES.LINE_SHAPES), self.stru_elements)

    @property
    def lines_hinged(self) -> Iterable[Elem]:
        return filter(lambda x: x.hinge_prop is not None, self.lines)

    @property
    def lines_ecc(self) -> Iterable[Elem]:
        return filter(lambda x: x.eccentricity is not None, self.lines)

    @property
    def connectors(self) -> Iterable[Connector]:
        return filter(lambda x: isinstance(x.type, Connector), self.elements)

    @property
    def masses(self) -> Iterable[Mass]:
        return filter(lambda x: isinstance(x, Mass), self.elements)

    @property
    def stru_elements(self) -> Iterable[Elem]:
        not_strus = (Mass, Connector)
        return filter(lambda x: isinstance(x, not_strus) is False, self._elements)

    def connector_by_name(self, name: str):
        """Get Connector by name"""
        cmap = {c.name: c for c in self.connectors}
        return cmap.get(name, None)

    def from_id(self, el_id: int) -> Union[Elem, Connector]:
        el = self._idmap.get(el_id, None)
        if el is None:
            spring_id_map = {m.id: m for m in self.parent.springs.values()}
            res = spring_id_map.get(el_id, None)
            if res is not None:
                return res

            raise ValueError(f'The elem id "{el_id}" is not found')
        return el

    def filter_elements(self, keep_elem=None, delete_elem=None):
        """
        Filter which element types you wish to keep using the "keep_elem" list of elem_types or which elements you wish
        to delete using the 'delete_elem' option.

        :param keep_elem:
        :param delete_elem:
        """
        from ada.fem.shapes.definitions import ShapeResolver

        keep_elem = [ShapeResolver.get_el_type_from_str(el_) for el_ in keep_elem] if keep_elem is not None else None
        delete_elem = (
            [ShapeResolver.get_el_type_from_str(el_) for el_ in delete_elem] if delete_elem is not None else None
        )

        def eval_elem(el):
            if keep_elem is not None:
                return True if el.type in keep_elem else False
            else:
                return False if el.type in delete_elem else True

        self._elements = list(filter(eval_elem, self._elements))
        self._by_types = dict(self.group_by_type())
        self._idmap = {e.id: e for e in self._elements}

    @property
    def idmap(self):
        return self._idmap

    def add(self, elem: Elem, skip_grouping=False) -> Elem:
        if elem.id is None:
            if len(self._elements) > 0:
                elem._el_id = self._elements[-1].id + 1
            else:
                elem._el_id = 1
        if elem.id in self.idmap.keys():
            raise ValueError(f'Elem id "{elem.id}" already exists or is not set.')

        if elem.parent is None:
            elem.parent = self._fem_obj

        self._elements.append(elem)
        self._idmap[elem.id] = elem

        if skip_grouping:
            return elem

        self._group_by_types()
        return elem

    def remove(self, elems: Union[Elem, List[Elem]]):
        """Remove elem or list of elements from container"""
        elems = list(elems) if isinstance(elems, Iterable) else [elems]
        for elem in elems:
            if elem in self._elements:
                logger.warning(f"Element removal is WIP. Removing element: {elem}")
                self._elements.pop(self._elements.index(elem))
            else:
                logger.error(f"'{elem}' not found in {self.__class__.__name__}-container.")
        # self._sort()

    def group_by_type(self):
        return groupby(sorted(self._elements, key=attrgetter("type")), key=attrgetter("type"))

    def _group_by_types(self):
        if len(self._elements) > 0:
            self._by_types = groupby(
                sorted(self._elements, key=attrgetter("type")), key=attrgetter("type", SetTypes.ELSET)
            )
        else:
            self._by_types = dict()

    def _sort(self):
        self._elements = sorted(self._elements, key=attrgetter("id"))
        self._group_by_types()
        self.renumber()

    def merge_with_coincident_nodes(self):
        def remove_duplicate_nodes():
            new_nodes = [n for n in elem.nodes if n.has_refs]
            elem.nodes.clear()
            elem.nodes.extend(new_nodes)

        """
        This does not work according to plan. It seems like it is deleting more and more from the model for each
        iteration
        """
        for elem in filter(lambda x: len(x.nodes) > len([n for n in x.nodes if n.has_refs]), self._elements):
            remove_duplicate_nodes()
            elem.update()

    def update(self):
        self.remove(list(filter(lambda x: x.id is None, self._elements)))


class FemSections:
    def __init__(self, sections: Iterable[FemSection] = None, fem_obj: "FEM" = None):
        self._fem_obj = fem_obj
        self._sections = list(sections) if sections is not None else []
        by_types = self._groupby()
        self._lines = by_types["lines"]
        self._shells = by_types["shells"]
        self._solids = by_types["solids"]
        self._name_map = {e.name: e for e in self._sections} if len(self._sections) > 0 else dict()
        self._id_map = {e.id: e for e in self._sections} if len(self._sections) > 0 else dict()
        if len(self._sections) > 0 and fem_obj is not None:
            self._link_data()

    def _map_by_properties(self) -> Dict[Tuple[Material, Section, tuple, tuple, float], List[FemSection]]:
        from ada import Material, Section

        merge_map: Dict[Tuple[Material, Section, tuple, tuple, float], List[FemSection]] = dict()
        for fs in self.lines:
            props = (fs.material, fs.section.unique_props(), tuple(), tuple(fs.local_z), tuple(fs.local_y))
            if props not in merge_map.keys():
                merge_map[props] = []

            merge_map[props].append(fs)

        for fs in self.shells:
            props = (fs.material, fs.section, (None,), tuple(), fs.thickness)
            if props not in merge_map.keys():
                merge_map[props] = []

            merge_map[props].append(fs)

        return merge_map

    def merge_by_properties(self):
        parent_part = self.parent.parent
        parent_part.move_all_mats_and_sec_here_from_subparts()

        prop_map = self._map_by_properties()
        remove_fs = []
        for _, fs_list in prop_map.items():
            fs_o = fs_list[0]
            el_o = fs_list[0].elset.members[0]
            for fs in fs_list[1:]:
                for el in fs.elset.members:
                    if el == el_o:
                        continue
                    are_equal = el.fem_sec.has_equal_props(fs_o)
                    if are_equal is True:
                        remove_fs.append(el.fem_sec)
                    el.fem_sec = fs_o
                    fs_o.elset.add_members([el])
        rest_list = list(set(remove_fs))
        self.remove(rest_list)

    def _map_materials(self, fem_sec: FemSection, mat_repo: Materials):
        if type(fem_sec.material) is str:
            logger.error(f'Material "{fem_sec.material}" was passed as string')
            fem_sec._material = mat_repo.get_by_name(fem_sec.material.name)

    def _map_elsets(self, fem_sec: FemSection, elset_repo):
        if type(fem_sec.elset) is str:
            if fem_sec.elset not in elset_repo:
                raise ValueError(f'The element set "{fem_sec.elset}" is not imported. ')
            fem_sec._elset = elset_repo[fem_sec.elset]
        elif type(fem_sec) is FemSection:
            pass
        else:
            raise ValueError("Invalid element set has been attached to this fem section")

    def _link_data(self):
        mat_repo = self._fem_obj.parent.get_assembly().materials
        list(map(partial(self._map_materials, mat_repo=mat_repo), self._sections))

        elsets_repo = self._fem_obj.elsets
        list(map(partial(self._map_elsets, elset_repo=elsets_repo), self._sections))
        [fem_sec.link_elements() for fem_sec in self._sections]

    def _groupby(self):
        return dict(
            lines=list(filter(lambda x: x.type == ElemType.LINE, self._sections)),
            shells=list(filter(lambda x: x.type == ElemType.SHELL, self._sections)),
            solids=list(filter(lambda x: x.type == ElemType.SOLID, self._sections)),
        )

    def __contains__(self, item: FemSection):
        return item in self._sections

    def __len__(self):
        return len(self._sections)

    def __iter__(self) -> Iterable[FemSection]:
        return iter(self._sections)

    def __getitem__(self, index):
        result = self._sections[index]
        return FemSections(result) if isinstance(index, slice) else result

    def __add__(self, other: FemSections):
        return FemSections(chain(self.sections, other.sections), fem_obj=self._fem_obj)

    def __repr__(self):
        return f"FemSections(Beams: {len(self.lines)}, Shells: {len(self.shells)}, Solids: {len(self.solids)})"

    @property
    def parent(self):
        return self._fem_obj

    @parent.setter
    def parent(self, value):
        self._fem_obj = value

    @property
    def sections(self) -> List[FemSection]:
        return self._sections

    @property
    def lines(self) -> List[FemSection]:
        return self._lines

    @property
    def shells(self) -> List[FemSection]:
        return self._shells

    @property
    def solids(self) -> List[FemSection]:
        return self._solids

    @property
    def name_map(self):
        return self._name_map

    @property
    def id_map(self):
        return self._id_map

    def index(self, item):
        index = bisect_left(self._sections, item)
        if (index != len(self._sections)) and (self._sections[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def _map_femsec_to_elem(self, elem: Elem, fem_sec: FemSection):
        elem.fem_sec = fem_sec

    def add(self, sec: FemSection):
        if sec.name in self.name_map.keys() or sec.name is None:
            raise ValueError(f'Section name "{sec.name}" already exists')
        if sec.id in self.id_map.keys() or sec.id is None:
            raise ValueError(f'Section ID "{sec.id}" already exists')

        self._sections.append(sec)
        if sec.type == ElemType.LINE:
            self._lines.append(sec)
        elif sec.type == ElemType.SHELL:
            self._shells.append(sec)
        else:
            self._solids.append(sec)

        [self._map_femsec_to_elem(el, fem_sec=sec) for el in sec.elset.members]
        self._name_map[sec.name] = sec
        self._id_map[sec.id] = sec

    def remove(self, fs_in: Union[List[FemSection], FemSection]):
        if type(fs_in) is not list:
            fs_in = [fs_in]

        for fs in fs_in:
            index = self._sections.index(fs)
            rem_fs = self._sections.pop(index)
            if len(rem_fs.elset.refs) == 1 and rem_fs.elset.refs[0] == rem_fs:
                rem_fs.parent.sets.remove(rem_fs.elset)

        self._name_map = {e.name: e for e in self._sections} if len(self._sections) > 0 else dict()
        self._id_map = {e.id: e for e in self._sections} if len(self._sections) > 0 else dict()

        by_types = self._groupby()
        self._lines = by_types["lines"]
        self._shells = by_types["shells"]
        self._solids = by_types["solids"]


class FemSets:
    def __init__(self, sets: List[FemSet] = None, parent: FEM = None):
        self._fem_obj = parent
        self._sets = sorted(sets, key=attrgetter("type", "name")) if sets is not None else []
        # Merge same name sets
        self._nomap = self._assemble_sets(self.is_nset) if len(self._sets) > 0 else dict()
        self._elmap = self._assemble_sets(self.is_elset) if len(self._sets) > 0 else dict()
        self._same_names = dict()
        if len(self._sets) > 0:
            self.link_data()

    def add_references(self) -> None:
        """Add reference to the containing FemSet for each member (node or element)"""

        def _map_ref(el, fem_set):
            el.refs.append(fem_set)

        for _set in self._sets:
            [_map_ref(m, _set) for m in _set.members]

    @property
    def parent(self) -> FEM:
        return self._fem_obj

    @parent.setter
    def parent(self, value):
        self._fem_obj = value

    @staticmethod
    def is_nset(fs):
        return True if fs.type == SetTypes.NSET else False

    @staticmethod
    def is_elset(fs):
        return True if fs.type == SetTypes.ELSET else False

    def _instantiate_all_members(self, fem_set: FemSet):
        from ada.fem import Connector, Mass, Spring

        def get_nset(nref):
            if type(nref) is Node:
                return nref
            else:
                return fem_set.parent.nodes.from_id(nref)

        def get_elset(elref):
            if type(elref) in (int, np.int32):
                return fem_set.parent.elements.from_id(elref)
            elif type(elref) is Elem:
                if elref not in elref.parent.elements and len(elref.parent.elements) != 0:
                    raise ValueError("Element might be doubly defined")
                else:
                    return elref
            elif type(elref) in (Spring, Mass, Connector):
                return elref
            else:
                raise ValueError(f"Elref type '{type(elref)}' is not recognized")

        def eval_set(fset):
            if fset.type == SetTypes.ELSET:
                el_type = Elem
                get_func = get_elset
            else:
                el_type = Node
                get_func = get_nset

            res = list(filter(lambda x: type(x) != el_type, fset.members))
            if len(res) > 0:
                fset._members = [get_func(m) for m in fset.members]

        if "generate" in fem_set.metadata.keys():
            if fem_set.metadata["generate"] is True and len(fem_set.members) == 0:
                gen_mem = fem_set.metadata["gen_mem"]
                fem_set._members = [i for i in range(gen_mem[0], gen_mem[1] + 1, gen_mem[2])]
                fem_set.metadata["generate"] = False

        if fem_set.type == SetTypes.NSET:
            if len(fem_set.members) == 1 and type(fem_set.members[0]) is str and type(fem_set.members[0]) is not Node:
                fem_set._members = self.nodes[fem_set.members[0]]
                fem_set.parent = self._fem_obj
                return fem_set

        eval_set(fem_set)
        fem_set.parent = self._fem_obj
        return fem_set

    def link_data(self):
        list(map(self._instantiate_all_members, self._sets))

    def _assemble_sets(self, set_type):
        elsets = dict()
        for elset in filter(set_type, self._sets):
            if elset.name not in elsets.keys():
                elsets[elset.name] = elset
            else:
                self._instantiate_all_members(elset)
                elsets[elset.name] += elset
        return elsets

    def __contains__(self, item):
        return item.name in self._elmap.keys()

    def __len__(self):
        return len(self._sets)

    def __iter__(self):
        return iter(self._sets)

    def __getitem__(self, index):
        result = self._sets[index]
        return FemSets(result, parent=self._fem_obj) if isinstance(index, slice) else result

    def __add__(self, other: FemSets):
        # TODO: make default choice for similar named sets in a global settings class
        for name, _set in other.nodes.items():
            _set.parent = self.parent
            if name in self._nomap.keys():
                logger.warning(f'Duplicate Node sets. Node set "{name}" exists')
            self.add(_set, merge_sets_if_duplicate=True)
        for name, _set in other.elements.items():
            _set.parent = self.parent
            if name in self._elmap.keys():
                logger.warning(f'Duplicate element sets. Element set "{name}" exists')
            self.add(_set, merge_sets_if_duplicate=True)
        return self

    def get_elset_from_name(self, name: str) -> FemSet:
        result = self._elmap.get(name, None)
        if result is None:
            raise ValueError(f'The element set "{name}" is not found')

        return result

    def get_nset_from_name(self, name: str) -> FemSet:
        lower_map = {key.lower(): value for key, value in self._nomap.items()}
        result = lower_map.get(name.lower(), None)
        if result is None:
            raise ValueError(f'The nodal set "{name}" is not found')

        return result

    @property
    def elements(self):
        return self._elmap

    @property
    def nodes(self):
        return self._nomap

    @property
    def sets(self):
        return self._sets

    def index(self, item):
        index = bisect_left(self._sets, item)
        if (index != len(self._sets)) and (self._sets[index] == item):
            return index
        raise ValueError(f"{repr(item)} not found")

    def remove(self, fe_set: FemSet):
        i = self._sets.index(fe_set)
        self._sets.pop(i)
        if fe_set.type == SetTypes.NSET:
            self._nomap.pop(fe_set.name)
            if fe_set.name in self._same_names.keys():
                self._same_names.pop(fe_set.name)
        else:
            self._elmap.pop(fe_set.name)
            if fe_set.name in self._same_names.keys():
                self._same_names.pop(fe_set.name)

        # To evalute if dependencies of set should be checked?
        # Against: This is a downstream object. FemSections would point to this set and remove during concatenation.

    def add(self, fe_set: FemSet, append_suffix_on_exist=False, merge_sets_if_duplicate=False) -> FemSet:
        if fe_set.type == SetTypes.NSET:
            if fe_set.name in self._nomap.keys():
                fem_set = self._nomap[fe_set.name]
                new_mem = [m for m in fe_set.members if m.id not in fem_set.members]
                fem_set.add_members(new_mem)
        else:
            if fe_set.name in self._elmap.keys():
                if append_suffix_on_exist is False and merge_sets_if_duplicate is False:
                    raise FemSetNameExists(fe_set.name)

                if merge_sets_if_duplicate is True:
                    o_set = self._elmap[fe_set.name]
                    for mem in fe_set.members:
                        if mem not in o_set.members:
                            o_set.members.append(mem)

                if fe_set.name not in self._same_names.keys():
                    self._same_names[fe_set.name] = 1
                else:
                    self._same_names[fe_set.name] += 1

                fe_set.name = f"{fe_set.name}_{self._same_names[fe_set.name]}"

        self.sets.append(fe_set)
        if fe_set.type == SetTypes.ELSET:
            self._elmap[fe_set.name] = fe_set
        else:
            self._nomap[fe_set.name] = fe_set

        if fe_set.parent is None:
            if self._fem_obj is None:
                raise ValueError("Fem obj cannot be none")
            fe_set.parent = self._fem_obj

        self._instantiate_all_members(fe_set)

        return fe_set



# GMSH Node numbering
# """
#         Line:                 Line3:          Line4:
#
#       v
#       ^
#       |
#       |
# 0-----+-----1 --> u   0----2----1     0---2---3---1
#
# """
# Note: Line3 is changed to 0-1-2 to comply with Abaqus node ordering
from ada.fem.shapes.definitions import ConnectorTypes, LineShapes, SpringTypes

line_edges = {
    LineShapes.LINE: [[0, 1]],
    LineShapes.LINE3: [[0, 2]],
    SpringTypes.SPRING2: [[0, 1]],
    ConnectorTypes.CONNECTOR: [[0, 1]],
}



from __future__ import annotations

import pathlib

from OCC.Core.BRep import BRep_Builder
from OCC.Core.Quantity import Quantity_Color, Quantity_TOC_RGB
from OCC.Core.STEPCAFControl import STEPCAFControl_Writer
from OCC.Core.STEPControl import STEPControl_AsIs
from OCC.Core.TCollection import TCollection_ExtendedString
from OCC.Core.TDataStd import TDataStd_Name
from OCC.Core.TDF import TDF_Label
from OCC.Core.TDocStd import TDocStd_Application, TDocStd_Document
from OCC.Core.TopoDS import TopoDS_Compound, TopoDS_Shape
from OCC.Core.XCAFDoc import XCAFDoc_ColorType, XCAFDoc_DocumentTool
from OCC.Core.XSControl import XSControl_WorkSession


class StepWriter:
    def __init__(self, top_level_name: str = "Assembly", schema: str = "AP242", assembly_mode: bool = True):
        app = TDocStd_Application()
        doc = TDocStd_Document(TCollection_ExtendedString("XmlOcaf"))
        app.InitDocument(doc)

        # The shape tool
        shape_tool = XCAFDoc_DocumentTool.ShapeTool(doc.Main())
        shape_tool.SetAutoNaming(False)

        # The color tool
        self.color_tool = XCAFDoc_DocumentTool.ColorTool(doc.Main())

        # Set up the compound
        comp = TopoDS_Compound()
        comp_builder = BRep_Builder()
        comp_builder.MakeCompound(comp)
        self.comp_builder = comp_builder

        self.doc = doc
        self.comp = comp

        top_level_label = shape_tool.AddShape(comp, False)
        set_name(top_level_label, top_level_name)
        self.tll = top_level_label
        self.shape_tool = shape_tool

    def add_shape(self, shape: TopoDS_Shape | TopoDS_Compound, name: str, rgb_color=None, parent=None):
        self.comp_builder.Add(self.comp, shape)
        parent = self.tll if parent is None else parent
        shape_label = self.shape_tool.AddSubShape(parent, shape)
        rgb_color = (1, 0, 0) if rgb_color is None else rgb_color
        set_color(shape_label, rgb_color, self.color_tool)
        set_name(shape_label, name)

    def export(self, step_file: pathlib.Path):
        # Set up the writer
        session = XSControl_WorkSession()
        writer = STEPCAFControl_Writer(session, False)
        writer.SetColorMode(True)
        writer.SetLayerMode(False)
        writer.SetNameMode(True)

        writer.Transfer(self.doc, STEPControl_AsIs)
        status = writer.Write(str(step_file))

        if not status:
            raise Exception("STEP export failed")
        else:
            print(f"STEP export status: {status}")


def set_name(label: TDF_Label, name: str):
    TDataStd_Name.Set(label, TCollection_ExtendedString(name))


def set_color(label: TDF_Label, color: tuple, tool):
    r, g, b = color
    tool.SetColor(label, Quantity_Color(r, g, b, Quantity_TOC_RGB), XCAFDoc_ColorType.XCAFDoc_ColorSurf)



from ada.config import get_logger
from ada.fem.shapes.definitions import LineShapes, ShellShapes, SolidShapes

logger = get_logger()


def ada_to_med_type(value):
    if value in _ada_to_med_type.keys():
        return _ada_to_med_type[value]
    else:
        for key, val in _ada_to_med_type.items():
            if type(key) is tuple:
                if value in key:
                    return val
    raise KeyError(f'Unsupported value "{value}"')


def med_to_ada_type(value):
    _tmp = {v: k for k, v in _ada_to_med_type.items()}

    if value not in _tmp:
        raise KeyError(f'Unsupported value "{value}"')

    res = _tmp[value]
    if type(res) is tuple:
        logger.info(f'Choosing index=0 -> "{res[0]}" when converting from MED type "{value}" to abaqus')
        return res[0]
    else:
        return res


_ada_to_med_type = {
    LineShapes.LINE: "SE2",
    LineShapes.LINE3: "SE3",
    ShellShapes.TRI: "TR3",
    ShellShapes.TRI6: "TR6",
    ShellShapes.TRI7: "TR7",  # Code Aster Specific type
    ShellShapes.QUAD: "QU4",
    ShellShapes.QUAD8: "QU8",
    ShellShapes.QUAD9: "QU9",  # Code Aster Specific type
    SolidShapes.TETRA: "TE4",
    SolidShapes.TETRA10: "T10",
    SolidShapes.HEX8: "HE8",
    SolidShapes.HEX20: "H20",
    SolidShapes.PYRAMID5: "PY5",
    # "pyramid13": "P13",
    SolidShapes.WEDGE: "PE6",
    # "wedge15": "P15",
}



from __future__ import annotations

import pathlib
import re
from dataclasses import dataclass, field
from enum import Enum
from typing import Iterator

import meshio
import numpy as np

from ada.base.types import BaseEnum
from ada.fem.results.common import ElementBlock, ElementInfo, FEAResult, FemNodes, Mesh
from ada.fem.results.field_data import (
    ElementFieldData,
    FieldData,
    NodalFieldData,
    NodalFieldType,
)


class ReadFrdFailedException(Exception):
    pass


class CcxPointData(BaseEnum):
    DISP = "DISP"
    FORC = "FORC"
    STRESS = "STRESS"
    PE = "PE"
    ERROR = "ERROR"


class CcxFieldData(BaseEnum):
    pass


class ElemShape(Enum):
    WEDGE = "wedge"
    WEDGE15 = "wedge15"
    HEX = "hexahedron"
    HEX20 = "hexahedron20"
    TETRA = "tetra"
    TETRA10 = "tetra10"

    @staticmethod
    def get_type_from_elem_array_shape(elements: np.ndarray) -> ElemShape:
        shape = elements.shape
        if shape[1] == 10:
            return ElemShape.WEDGE
        elif shape[1] == 12:
            return ElemShape.HEX
        elif shape[1] == 8:
            return ElemShape.TETRA
        elif shape[1] == 24:
            return ElemShape.HEX20
        elif shape[1] == 19:
            return ElemShape.WEDGE15
        elif shape[1] == 14:
            return ElemShape.TETRA10
        else:
            raise NotImplementedError(f"{shape=}")

    @staticmethod
    def el_shape_to_baseshape(shape: ElemShape):
        from ada.fem.shapes.definitions import SolidShapes

        if shape == ElemShape.HEX:
            return SolidShapes.HEX8
        elif shape == ElemShape.HEX20:
            return SolidShapes.HEX20
        elif shape == ElemShape.WEDGE:
            return SolidShapes.WEDGE
        elif shape == ElemShape.TETRA:
            return SolidShapes.TETRA
        elif shape == ElemShape.WEDGE15:
            return SolidShapes.WEDGE15
        elif shape == ElemShape.TETRA10:
            return SolidShapes.TETRA10
        else:
            raise NotImplementedError(f"{shape=}")


@dataclass
class CcxResultModel:
    file: Iterator

    ccx_version: str = None
    nodes: np.ndarray = None
    elements: np.ndarray = None
    results: list[NodalFieldData | ElementFieldData] = field(default_factory=list)

    _curr_step: int = None
    _curr_mode: int = None
    _curr_eig_freq: float = None
    _curr_eig_value: float = None

    def collect_nodes(self):
        while True:
            data = next(self.file)
            stripped = data.strip()
            if stripped.startswith("-1") is False:
                break
            split = stripped.split()
            try:
                yield [float(x) for x in split[1:]]
            except ValueError:  # Typically it's a - character separating two columns not whitespace
                split = list(filter(lambda x: x.strip() != "", re.split(r"(?=\s|(?<=[0-9])-(?=[0-9]))", stripped)))
                yield [float(x) for x in split[1:]]

    def collect_elements(self):
        elements = []
        curr_element = []

        while True:
            data = next(self.file)

            stripped = data.strip()
            if stripped.startswith("-1") is False:
                if stripped.startswith("-2") is False:
                    break

            split = stripped.split()

            if stripped.startswith("-1"):
                if len(curr_element) != 0:
                    elements.append(tuple(curr_element))
                data = [int(x) for x in split[1:]]
                curr_element = data
            elif stripped.startswith("-2"):
                data = [int(x) for x in split[1:]]
                curr_element += data
            else:
                break

        self.elements = np.asarray(elements)
        self.eval_flags(data)

    def collect_results(self, first_line):
        name = first_line.split()[1]

        component_names = []
        component_data = []
        while True:
            data = next(self.file)
            stripped = data.strip()

            if stripped.startswith("-5"):
                split = stripped.split()
                component_names.append(split[1])
            elif stripped.startswith("-1"):
                slen = len(stripped)
                x = 12
                res = [stripped[i : i + x] for i in range(0, slen, x)]
                first_batch = res[0].split()
                nid = int(float(first_batch[1]))
                component_data.append((nid, *[float(x) for x in res[1:]]))
            else:
                break

        curr_step = self._curr_step
        curr_eig_freq = None
        curr_eig_value = None

        if self._curr_mode is not None:
            curr_step = self._curr_mode
            curr_eig_value = self._curr_eig_value
            curr_eig_freq = self._curr_eig_freq

        field_type = None
        if name.startswith("DISP"):
            field_type = NodalFieldType.DISP

        self.results.append(
            NodalFieldData(
                name,
                curr_step,
                component_names,
                np.asarray(component_data),
                eigen_freq=curr_eig_freq,
                eigen_value=curr_eig_value,
                field_type=field_type,
            )
        )
        self.eval_flags(data)

    def eval_flags(self, data: str):
        stripped = data.strip()
        if stripped.startswith("1UVERSION"):
            res = stripped.split()
            self.ccx_version = res[-1].lower().replace("version", "").strip()

        if stripped.startswith("2C"):
            split = stripped.split()
            num_len = int(float(split[1]))
            # Note! np.fromiter have issues on older numpy versions such as (1.22.3)
            self.nodes = np.fromiter(self.collect_nodes(), dtype=np.dtype((float, 4)), count=num_len)

        if stripped.startswith("3C"):
            self.collect_elements()

        if stripped.startswith("1PSTEP"):
            split_data = stripped.split()
            self._curr_step = int(float(split_data[2]))

        if stripped.startswith("1PMODE"):
            split_data = stripped.split()
            self._curr_mode = int(float(split_data[-1]))

        if stripped.startswith("1PGK"):
            split_data = stripped.split()
            self._curr_eig_value = float(split_data[1])

        if stripped.startswith("100CL"):
            split_data = stripped.split()
            self._curr_eig_freq = float(split_data[2])

        if stripped.startswith("-4"):
            self.collect_results(stripped)

    def load(self):
        while True:
            try:
                curr = next(self.file)
                self.eval_flags(curr)
            except StopIteration:
                break

    def get_last_step_results(self, data_type: BaseEnum) -> list[FieldData]:
        results = dict()
        for res in self.results:
            if data_type.from_str(res.name) is None:
                continue
            existing_step = results.get(res.name)
            if existing_step is not None:
                if res.step > existing_step.step:
                    results[res.name] = res
            else:
                results[res.name] = res

        return list(results.values())


def read_from_frd_file(frd_file) -> meshio.Mesh:
    with open(frd_file, "r") as f:
        ccx_res_model = CcxResultModel(f)
        ccx_res_model.load()

    mesh = to_meshio_mesh(ccx_res_model)
    return mesh


def read_from_frd_file_proto(frd_file) -> FEAResult:
    with open(frd_file, "r", encoding="utf8") as f:
        ccx_res_model = CcxResultModel(f)
        ccx_res_model.load()

    if ccx_res_model.elements is None:
        raise ReadFrdFailedException("No element information from Calculix")

    return to_fea_result_obj(ccx_res_model, frd_file)


def to_fea_result_obj(ccx_results: CcxResultModel, frd_file) -> FEAResult:
    from ada.fem.formats.general import FEATypes

    if isinstance(frd_file, str):
        frd_file = pathlib.Path(frd_file)

    description = f"Adapy - Calculix ({ccx_results.ccx_version}) Results"
    shape = ElemShape.get_type_from_elem_array_shape(ccx_results.elements)
    node_refs = ccx_results.elements[:, 4:]
    elem_info = ElementInfo(
        type=ElemShape.el_shape_to_baseshape(shape), source_software=FEATypes.CALCULIX, source_type=str(shape.value)
    )
    identifiers = ccx_results.elements[:, 0]
    elem_block = ElementBlock(elem_info=elem_info, node_refs=node_refs, identifiers=identifiers)

    coords = ccx_results.nodes[:, 1:]
    identifiers = ccx_results.nodes[:, 0]
    nodes = FemNodes(coords=coords, identifiers=identifiers)
    mesh = Mesh(elements=[elem_block], nodes=nodes)

    return FEAResult(
        frd_file.name,
        FEATypes.CALCULIX,
        ccx_results.results,
        mesh=mesh,
        results_file_path=frd_file,
        description=description,
    )


def to_meshio_mesh(ccx_results: CcxResultModel) -> meshio.Mesh:
    # Points
    nodes = ccx_results.nodes
    if nodes is None:
        raise ReadFrdFailedException("No nodes found. Maybe there was an issue with the analysis")

    points = nodes[:, 1:]
    monotonic_point_map = dict()
    for i, x in enumerate(nodes[:, 0].astype(int), start=0):
        monotonic_point_map[x] = i

    # Cells
    elements = ccx_results.elements
    cells = elements[:, 4:]
    for original_num, new_num in monotonic_point_map.items():
        cells[cells == original_num] = new_num

    shape = ElemShape.get_type_from_elem_array_shape(elements)
    cell_block = meshio.CellBlock(str(shape.value), cells)

    # Point Data
    # Multiple steps are AFAIK not supported in the meshio Mesh object. So only the last step is used
    point_data = dict()
    for res in ccx_results.get_last_step_results(CcxPointData):
        values = np.asarray(res.values)[:, 1:]
        point_data[res.name] = values

    # Field Data
    cell_data = dict()
    for res in ccx_results.get_last_step_results(CcxFieldData):
        values = np.asarray(res.values)[:, 1:]
        cell_data[res.name] = [values]

    mesh = meshio.Mesh(points=points, cells=[cell_block], cell_data=cell_data, point_data=point_data)
    return mesh


def safesplit(stripped):
    return list(filter(lambda x: x.strip() != "", re.split(r"(?=\s|(?<=[0-9])-(?=[0-9]))", stripped)))



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Union

from ada.config import get_logger

from .common import FemBase
from .constraints import Bc
from .interactions import Interaction
from .loads import Load, LoadCase, LoadGravity, LoadPressure
from .outputs import FieldOutput, HistOutput

if TYPE_CHECKING:
    from ada import FEM

logger = get_logger()


class _StepTypes:
    STATIC = "static"
    EIGEN = "eigenfrequency"
    COMPLEX_EIG = "complex_eig"
    STEADY_STATE = "steady_state"
    DYNAMIC = "dynamic"
    EXPLICIT = "explicit"

    all = [STATIC, EIGEN, STEADY_STATE, DYNAMIC, COMPLEX_EIG, EXPLICIT]


class _DynStepType:
    QUASI_STATIC = "QUASI-STATIC"
    TRANSIENT_FIDELITY = "TRANSIENT FIDELITY"
    MODERATE_DISSIPATION = "MODERATE DISSIPATION"
    all = [QUASI_STATIC, TRANSIENT_FIDELITY, MODERATE_DISSIPATION]


class StepSolverOptions:
    """A class for FE solver specific options"""

    def __init__(self):
        self._ABAQUS = None
        self._CODE_ASTER = None
        self._CALCULIX = None

    @property
    def CODE_ASTER(self):
        return NotImplementedError()

    @property
    def CALCULIX(self):
        return NotImplementedError()

    @property
    def ABAQUS(self):
        if self._ABAQUS is None:
            from .formats.abaqus.solver import AbaqusStepOptions

            self._ABAQUS = AbaqusStepOptions()
        return self._ABAQUS


class Step(FemBase):
    """
    A FEM analysis step object

    :param name: Name of step
    :param step_type: Step type: | 'static' | 'eigenfrequency' |  'response_analysis' | 'dynamic' | 'complex_eig' |
    :param nl_geom: Include or ignore the nonlinear effects of large deformations and displacements (default=False)
    """

    TYPES = _StepTypes

    def __init__(
        self,
        name,
        step_type,
        nl_geom=False,
        total_time=None,
        solver_options: StepSolverOptions = StepSolverOptions(),
        use_default_outputs=True,
        metadata=None,
        parent: "FEM" = None,
    ):
        super().__init__(name, metadata, parent)
        if step_type not in _StepTypes.all:
            raise ValueError(f'Step type "{step_type}" is currently not supported')

        self._total_time = total_time
        self._step_type = step_type
        self._nl_geom = nl_geom
        self._solver_options = solver_options
        self._bcs: Dict[str, Bc] = dict()
        self._loads: List[Load] = []
        self._load_cases = dict()
        self._interactions = dict()
        self._hist_outputs = []
        self._field_outputs = []

        if use_default_outputs:
            hist, field = self.get_default_output_variables()
            self._hist_outputs += [hist]
            self._field_outputs += [field]

    def get_default_output_variables(self):
        from ada.fem.outputs import defaults

        return defaults()

    def add_load(self, load: Union[Load, LoadPressure, LoadGravity]):
        if isinstance(load, LoadPressure):
            if load.surface.parent is None:
                self.parent.add_surface(load.surface)
        load.parent = self
        self._loads.append(load)

    def add_loadcase(self, load_case: LoadCase):
        for load in load_case.loads:
            if load not in self.loads:
                self.loads.append(load)
                load.parent = self
        load_case.parent = self
        self._load_cases[load_case.name] = load_case

    def add_bc(self, bc: Bc):
        bc.parent = self
        self._bcs[bc.name] = bc
        if bc.fem_set.parent is None and bc.fem_set not in self.parent.sets:
            self.parent.sets.add(bc.fem_set)
        if bc.amplitude is not None and bc.amplitude.parent is None:
            self.parent.add_amplitude(bc.amplitude)

    def add_history_output(self, hist_output: HistOutput):
        hist_output.parent = self
        if hist_output.fem_set.parent is None and self.parent is not None:
            self.parent.add_set(hist_output.fem_set)
        self._hist_outputs.append(hist_output)

    def add_field_output(self, field_output: FieldOutput):
        field_output.parent = self
        self._field_outputs.append(field_output)

    def add_interaction(self, interaction: Interaction):
        interaction.parent = self
        self._interactions[interaction.name] = interaction

    @property
    def type(self):
        return self._step_type

    @property
    def nl_geom(self):
        return self._nl_geom

    @property
    def options(self) -> StepSolverOptions:
        return self._solver_options

    @options.setter
    def options(self, value):
        self._solver_options = value

    @property
    def total_time(self):
        return self._total_time

    @property
    def interactions(self) -> Dict[str, Interaction]:
        return self._interactions

    @property
    def bcs(self) -> Dict[str, Bc]:
        return self._bcs

    @property
    def loads(self) -> List[Load]:
        return self._loads

    @property
    def load_cases(self) -> Dict[str, LoadCase]:
        return self._load_cases

    @property
    def field_outputs(self) -> List[FieldOutput]:
        return self._field_outputs

    @property
    def hist_outputs(self) -> List[HistOutput]:
        return self._hist_outputs

    def __repr__(self):
        return f"{self.__class__.name}({self.name}, type={self.type}, nl_geom={self.nl_geom})"


class StepImplicit(Step):
    TYPES_DYNAMIC = _DynStepType

    def __init__(
        self,
        name,
        implicit_type=Step.TYPES.STATIC,
        nl_geom=False,
        total_time=100.0,
        total_incr=1000,
        init_incr=100.0,
        min_incr=1e-8,
        max_incr=100.0,
        dyn_type=TYPES_DYNAMIC.QUASI_STATIC,
        **kwargs,
    ):
        """
        :param total_incr: Maximum number of allowed increments
        :param init_incr: Initial increment
        :param total_time: Total step time
        :param min_incr: Minimum allowable increment size
        :param max_incr: Maximum allowable increment size
        :param dyn_type: Dynamic analysis type 'TRANSIENT FIDELITY' | 'QUASI-STATIC'
        """
        if total_time is not None:
            if init_incr > total_time and nl_geom is True:
                logger.warning(
                    f"Initial increment > Total time ({init_incr} > {total_time}). "
                    "Adjusted initial increment equal to total time"
                )
                init_incr = total_time
        else:
            total_time = init_incr

        super(StepImplicit, self).__init__(name, implicit_type, total_time=total_time, nl_geom=nl_geom, **kwargs)

        if dyn_type not in _DynStepType.all:
            raise ValueError(f'Dynamic input type "{dyn_type}" is not supported')

        self._dyn_type = dyn_type
        self._total_incr = total_incr
        self._init_incr = init_incr
        self._min_incr = min_incr
        self._max_incr = max_incr

    @property
    def total_incr(self):
        return self._total_incr

    @property
    def init_incr(self):
        return self._init_incr

    @property
    def min_incr(self):
        return self._min_incr

    @property
    def max_incr(self):
        return self._max_incr

    @property
    def dyn_type(self):
        return self._dyn_type


class StepExplicit(Step):
    def __init__(self, name, **kwargs):
        super(StepExplicit, self).__init__(name, Step.TYPES.EXPLICIT, **kwargs)


class StepEigen(Step):
    def __init__(self, name, num_eigen_modes: int, **kwargs):
        super(StepEigen, self).__init__(name, Step.TYPES.EIGEN, **kwargs)
        self._num_eigen_modes = num_eigen_modes

    @property
    def num_eigen_modes(self):
        """Number of requested Eigen modes"""
        return self._num_eigen_modes


class StepEigenComplex(StepEigen):
    def __init__(self, name, num_eigen_modes, friction_damping=False, **kwargs):
        super(StepEigenComplex, self).__init__(name, num_eigen_modes, **kwargs)
        self._friction_damping = friction_damping

    @property
    def friction_damping(self):
        return self._friction_damping


class StepSteadyState(Step):
    """
    :param alpha: Rayleigh Damping
    :param beta: Rayleigh Damping
    :param unit_load: Unit Load
    :param fmin: Minimum frequency
    :param fmax: Maximum frequency
    """

    def __init__(self, name, unit_load: Load, fmin, fmax, alpha=0.1, beta=10, **kwargs):
        super(StepSteadyState, self).__init__(name, Step.TYPES.STEADY_STATE, **kwargs)
        self._alpha = alpha
        self._beta = beta
        self._fmin = fmin
        self._fmax = fmax
        self._unit_load = unit_load

    @property
    def fmin(self):
        return self._fmin

    @property
    def fmax(self):
        return self._fmax

    @property
    def alpha(self):
        return self._alpha

    @property
    def beta(self):
        return self._beta

    @property
    def unit_load(self):
        return self._unit_load



import xml.etree.ElementTree as ET

from ada.config import get_logger
from ada.fem.formats.sesam.xml.read.read_beams import el_to_beam
from ada.fem.formats.sesam.xml.read.read_materials import get_materials
from ada.fem.formats.sesam.xml.read.read_sections import get_sections

logger = get_logger()


def iter_beams_from_xml(xml_path):
    from ada import Part

    xml_root = ET.parse(str(xml_path)).getroot()
    all_beams = xml_root.findall(".//straight_beam") + xml_root.findall(".//curved_beam")
    p = Part("tmp")
    p._sections = get_sections(xml_root, p)
    p._materials = get_materials(xml_root, p)
    for bm_el in all_beams:
        yield from el_to_beam(bm_el, p)


def apply_mass_density_factors(root, p):
    mass_density_factors = {e.attrib["name"]: float(e.attrib["factor"]) for e in root.findall(".//mass_density_factor")}
    for bm in p.beams:
        mdf = bm.metadata.get("mass_density_factor_ref", None)
        if mdf is None:
            continue

        mdf_value = mass_density_factors[mdf]
        mat_name = f"{bm.material.name}_{mdf}"
        existing_mat = p.materials.name_map.get(mat_name, None)

        if existing_mat is None:
            bm.material = bm.material.copy_to(new_name=mat_name)
            bm.material.model.rho *= mdf_value
            p.add_material(bm.material)
        else:
            bm.material = existing_mat


def yield_plate_elems_to_plate(plate_elem, parent, sat_ref_d, thick_map):
    from ada import Plate

    mat = parent.materials.get_by_name(plate_elem.attrib["material_ref"])
    for i, res in enumerate(plate_elem.findall(".//face"), start=1):
        face_ref = res.attrib["face_ref"]
        points = sat_ref_d.get(face_ref, None)
        if points is None:
            logger.warning(f'Unable to find face_ref="{face_ref}"')
            continue

        name = plate_elem.attrib["name"]
        if i > 1:
            name += f"_{i:02d}"

        t = thick_map.get(plate_elem.attrib["thickness_ref"])
        try:
            pl = Plate(
                name,
                points,
                t,
                mat=mat,
                metadata=dict(props=dict(gxml_face_ref=face_ref)),
                use3dnodes=True,
                parent=parent,
            )
        except BaseException as e:
            logger.error(f"Failed converting plate {name} due to {e}")
            continue
        yield pl



from __future__ import annotations

import pathlib
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Iterable

import numpy as np

from ada.base.ifc_types import ShapeTypes
from ada.base.physical_objects import BackendGeom
from ada.base.units import Units
from ada.core.utils import Counter, roundoff
from ada.core.vector_utils import unit_vector, vector_length
from ada.materials import Material
from ada.materials.utils import get_material

from .bounding_box import BoundingBox
from .curves import CurvePoly
from .transforms import Placement

if TYPE_CHECKING:
    from OCC.Core.TopoDS import TopoDS_Shape

    from ada.ifc.store import IfcStore


class Shape(BackendGeom):
    IFC_CLASSES = ShapeTypes

    def __init__(
        self,
        name,
        geom,
        colour=None,
        opacity=1.0,
        mass: float = None,
        cog: tuple[float, float, float] = None,
        metadata=None,
        units=Units.M,
        guid=None,
        material: Material | str = None,
        placement=Placement(),
        ifc_store: IfcStore = None,
        ifc_class: ShapeTypes = ShapeTypes.IfcBuildingElementProxy,
    ):
        super().__init__(
            name,
            guid=guid,
            metadata=metadata,
            units=units,
            placement=placement,
            ifc_store=ifc_store,
            colour=colour,
            opacity=opacity,
        )
        if type(geom) in (str, pathlib.WindowsPath, pathlib.PurePath, pathlib.Path):
            from OCC.Extend.DataExchange import read_step_file

            geom = read_step_file(str(geom))

        self._geom = geom
        self._mass = mass
        self._cog = cog
        if isinstance(material, Material):
            self._material = material
        else:
            self._material = get_material(material)

        self._bbox = None
        self._ifc_class = ifc_class

    @property
    def type(self):
        return type(self.geom())

    @property
    def mass(self) -> float:
        return self._mass

    @mass.setter
    def mass(self, value: float):
        self._mass = value

    @property
    def cog(self) -> tuple[float, float, float]:
        return self._cog

    @cog.setter
    def cog(self, value: tuple[float, float, float]):
        self._cog = value

    @property
    def bbox(self) -> BoundingBox:
        if self._bbox is None and self.geom() is not None:
            self._bbox = BoundingBox(self)

        return self._bbox

    @property
    def point_on(self):
        return self.bbox[3:6]

    def geom(self) -> TopoDS_Shape:
        from ada.occ.utils import apply_penetrations

        from .exceptions import NoGeomPassedToShapeError

        if self._geom is None:
            from ada.ifc.read.read_shapes import get_ifc_geometry

            a = self.get_assembly()
            if a.ifc_store is not None:
                ifc_elem = a.ifc_store.get_by_guid(self.guid)
            else:
                raise NoGeomPassedToShapeError(f'No geometry information attached to shape "{self}"')

            geom, color, alpha = get_ifc_geometry(ifc_elem, a.ifc_store.settings)
            self._geom = geom
            self.colour = color
            self.opacity = alpha

        geom = apply_penetrations(self._geom, self.penetrations)

        return geom

    def solid(self) -> TopoDS_Shape:
        return self.geom()

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            scale_factor = Units.get_scale_factor(self._units, value)
            if self._geom is not None:
                from ada.occ.utils import transform_shape

                self._geom = transform_shape(self.geom(), scale_factor)

            if self.metadata.get("ifc_source") is True:
                raise NotImplementedError()

            self._units = value

    @property
    def material(self) -> Material:
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def ifc_class(self) -> ShapeTypes:
        return self._ifc_class

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.name}")'


class PrimSphere(Shape):
    def __init__(self, name, cog, radius, **kwargs):
        self.radius = radius
        super(PrimSphere, self).__init__(name=name, geom=None, cog=cog, **kwargs)

    def geom(self):
        from ada.occ.utils import apply_penetrations

        if self._geom is None:
            from ada.occ.utils import make_sphere

            self._geom = make_sphere(self.cog, self.radius)

        geom = apply_penetrations(self._geom, self.penetrations)
        return geom

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            from ada.occ.utils import make_sphere

            scale_factor = Units.get_scale_factor(self._units, value)
            self.cog = tuple([x * scale_factor for x in self.cog])
            self.radius = self.radius * scale_factor
            self._geom = make_sphere(self.cog, self.radius)
            self._units = value

    def __repr__(self):
        return f"PrimSphere({self.name})"


class PrimBox(Shape):
    """Primitive Box. Length, width & height are local x, y and z respectively"""

    def __init__(self, name, p1, p2, **kwargs):
        from ada.occ.utils import make_box_by_points

        self.p1 = p1
        self.p2 = p2
        super(PrimBox, self).__init__(name=name, geom=make_box_by_points(p1, p2), **kwargs)
        self._bbox = BoundingBox(self)

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            from ada.occ.utils import make_box_by_points

            scale_factor = Units.get_scale_factor(self._units, value)
            self.p1 = tuple([x * scale_factor for x in self.p1])
            self.p2 = tuple([x * scale_factor for x in self.p2])
            self._geom = make_box_by_points(self.p1, self.p2)
            self._units = value

    def __repr__(self):
        return f"PrimBox({self.name})"


class PrimCyl(Shape):
    def __init__(self, name, p1, p2, r, **kwargs):
        from ada.occ.utils import make_cylinder_from_points

        self.p1 = np.array(p1)
        self.p2 = np.array(p2)
        self.r = r
        super(PrimCyl, self).__init__(name, make_cylinder_from_points(p1, p2, r), **kwargs)

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        from ada.occ.utils import make_cylinder_from_points

        if value != self._units:
            scale_factor = Units.get_scale_factor(self._units, value)
            self.p1 = [x * scale_factor for x in self.p1]
            self.p2 = [x * scale_factor for x in self.p2]
            self.r = self.r * scale_factor
            self._geom = make_cylinder_from_points(self.p1, self.p2, self.r)

    def __repr__(self):
        return f"PrimCyl({self.name})"


class PrimExtrude(Shape):
    def __init__(self, name, curve: list[tuple], h, normal=None, origin=None, xdir=None, tol=1e-3, **kwargs):
        self._name = name

        poly = CurvePoly(
            points2d=curve,
            normal=normal,
            origin=origin,
            xdir=xdir,
            tol=tol,
            parent=self,
        )

        self._poly = poly
        self._extrude_depth = h

        super(PrimExtrude, self).__init__(name, self._poly.make_extruded_solid(self._extrude_depth), **kwargs)

    @staticmethod
    def from_2points_and_curve(name: str, p1: Iterable, p2: Iterable, profile: list[tuple], xdir: tuple) -> PrimExtrude:
        p1 = np.array(p1)
        p2 = np.array(p2)
        normal = unit_vector(p2 - p1)
        length = vector_length(p2 - p1)
        return PrimExtrude(name=name, curve=profile, h=length, normal=normal, origin=p1, xdir=xdir)

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            from ada.config import Settings

            scale_factor = Units.get_scale_factor(self._units, value)
            tol = Settings.mmtol if value == "mm" else Settings.mtol
            self.poly.scale(scale_factor, tol)
            self._extrude_depth = self._extrude_depth * scale_factor
            self._units = value

    @property
    def poly(self) -> CurvePoly:
        return self._poly

    @property
    def extrude_depth(self):
        return self._extrude_depth

    def __repr__(self):
        return f"PrimExtrude({self.name})"


class PrimRevolve(Shape):
    """Revolved Primitive"""

    def __init__(self, name, points2d, origin, xdir, normal, rev_angle, tol=1e-3, **kwargs):
        self._name = name
        poly = CurvePoly(
            points2d=points2d,
            normal=[roundoff(x) for x in normal],
            origin=origin,
            xdir=[roundoff(x) for x in xdir],
            tol=tol,
            parent=self,
        )
        self._poly = poly
        self._revolve_angle = rev_angle
        self._revolve_axis = [roundoff(x) for x in poly.ydir]
        self._revolve_origin = origin
        super(PrimRevolve, self).__init__(
            name,
            self._poly.make_revolve_solid(
                self._revolve_axis,
                self._revolve_angle,
                self._revolve_origin,
            ),
            **kwargs,
        )

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            from ada.config import Settings

            scale_factor = Units.get_scale_factor(self._units, value)
            tol = Settings.mmtol if value == "mm" else Settings.mtol
            self.poly.scale(scale_factor, tol)
            self._revolve_origin = [x * scale_factor for x in self.revolve_origin]
            self._geom = self._poly.make_revolve_solid(
                self._revolve_axis,
                self._revolve_angle,
                self._revolve_origin,
            )

    @property
    def poly(self) -> CurvePoly:
        return self._poly

    @property
    def revolve_origin(self):
        return self._revolve_origin

    @property
    def revolve_axis(self):
        return self._revolve_axis

    @property
    def revolve_angle(self):
        return self._revolve_angle

    def __repr__(self):
        return f"PrimRevolve({self.name})"


class PrimSweep(Shape):
    def __init__(
        self,
        name,
        sweep_curve,
        normal,
        xdir,
        profile_curve_outer,
        profile_curve_inner=None,
        origin=None,
        tol=1e-3,
        **kwargs,
    ):
        if type(sweep_curve) is list:
            sweep_curve = CurvePoly(points3d=sweep_curve, is_closed=False)

        if type(profile_curve_outer) is list:
            origin = sweep_curve.placement.origin if origin is None else origin
            profile_curve_outer = CurvePoly(profile_curve_outer, origin=origin, normal=normal, xdir=xdir)

        sweep_curve.parent = self
        profile_curve_outer.parent = self

        self._sweep_curve = sweep_curve
        self._profile_curve_outer = profile_curve_outer
        self._profile_curve_inner = profile_curve_inner

        super(PrimSweep, self).__init__(name, self._sweep_geom(), **kwargs)

    def _sweep_geom(self):
        from ada.occ.utils import sweep_geom

        return sweep_geom(self.sweep_curve.wire, self.profile_curve_outer.wire)

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            raise NotImplementedError()

    @property
    def sweep_curve(self):
        return self._sweep_curve

    @property
    def profile_curve_outer(self):
        return self._profile_curve_outer

    @property
    def profile_curve_inner(self):
        return self._profile_curve_inner

    def __repr__(self):
        return f"PrimSweep({self.name})"


class IfcBSplineSurfaceForm(Enum):
    PLANE_SURF = "PLANE_SURF"
    CYLINDRICAL_SURF = "CYLINDRICAL_SURF"
    CONICAL_SURF = "CONICAL_SURF"
    SPHERICAL_SURF = "SPHERICAL_SURF"
    TOROIDAL_SURF = "TOROIDAL_SURF"
    SURF_OF_REVOLUTION = "SURF_OF_REVOLUTION"
    RULED_SURF = "RULED_SURF"
    GENERALISED_CONE = "GENERALISED_CONE"
    QUADRIC_SURF = "QUADRIC_SURF"
    SURF_OF_LINEAR_EXTRUSION = "SURF_OF_LINEAR_EXTRUSION"
    UNSPECIFIED = "UNSPECIFIED"


@dataclass
class BSplineSurfaceWithKnots:
    """https://standards.buildingsmart.org/IFC/RELEASE/IFC4_3/lexical/IfcBSplineSurfaceWithKnots.htm"""

    uDegree: int
    vDegree: int
    controlPointsList: list[list[float]]
    surfaceForm: IfcBSplineSurfaceForm
    uKnots: list[float]
    vKnots: list[float]
    uMultiplicities: list[int]
    vMultiplicities: list[int]

    def get_entities(self):
        return {
            "UDegree": self.uDegree,
            "VDegree": self.vDegree,
            "ControlPointsList": self.controlPointsList,
            "SurfaceForm": self.surfaceForm.value,
            "UClosed": False,
            "VClosed": False,
            "SelfIntersect": False,
            "UKnots": self.uKnots,
            "VKnots": self.vKnots,
            "UMultiplicities": self.uMultiplicities,
            "VMultiplicities": self.vMultiplicities,
            "KnotSpec": "UNSPECIFIED",
        }

    def to_ifcopenshell(self, f):
        from ada.ifc.utils import ifc_p

        entities = self.get_entities()
        entities["ControlPointsList"] = [[ifc_p(f, i) for i in x] for x in self.controlPointsList]
        return f.create_entity("IfcBSplineSurfaceWithKnots", **entities)


@dataclass
class RationalBSplineSurfaceWithKnots(BSplineSurfaceWithKnots):
    """https://standards.buildingsmart.org/IFC/RELEASE/IFC4_3/lexical/IfcRationalBSplineSurfaceWithKnots.htm"""

    weightsData: list[list[float]]

    def get_entities(self):
        entities = super().get_entities()
        entities["WeightsData"] = self.weightsData
        return entities

    def to_ifcopenshell(self, f):
        from ada.ifc.utils import ifc_p

        entities = self.get_entities()
        entities["ControlPointsList"] = [[ifc_p(f, i[:3]) for i in x] for x in self.controlPointsList]
        return f.create_entity("IfcRationalBSplineSurfaceWithKnots", **entities)


class Penetration(BackendGeom):
    _name_gen = Counter(1, "Pen")
    """A penetration object. Wraps around a primitive"""

    # TODO: Maybe this class should be evaluated for removal?
    def __init__(self, primitive, metadata=None, parent=None, units=Units.M, guid=None):
        if issubclass(type(primitive), Shape) is False:
            raise ValueError(f'Unsupported primitive type "{type(primitive)}"')

        super(Penetration, self).__init__(primitive.name, guid=guid, metadata=metadata, units=units)
        self._primitive = primitive
        self._parent = parent
        self._ifc_opening = None

    @property
    def primitive(self):
        return self._primitive

    def geom(self):
        return self.primitive.geom()

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if value != self._units:
            self.primitive.units = value
            self._units = value

    def __repr__(self):
        return f"Pen(type={self.primitive})"



from __future__ import annotations

from typing import TYPE_CHECKING, List, Tuple, Union

import ifcopenshell
import ifcopenshell.geom
import ifcopenshell.util.element
import numpy as np
from ifcopenshell.util.unit import get_prefix_multiplier

import ada.core.constants as ifco
from ada.concepts.transforms import Transform
from ada.config import get_logger
from ada.core.file_system import get_list_of_files

if TYPE_CHECKING:
    from ada import Assembly, Beam

logger = get_logger()


def create_reference_subrep(f, global_axes):
    model_rep = f.create_entity("IfcGeometricRepresentationContext", None, "Model", 3, 1.0e-05, global_axes, None)
    body_sub_rep = f.create_entity(
        "IfcGeometricRepresentationSubContext",
        "Body",
        "Model",
        None,
        None,
        None,
        None,
        model_rep,
        None,
        "MODEL_VIEW",
        None,
    )
    ref_sub_rep = f.create_entity(
        "IfcGeometricRepresentationSubContext",
        "Reference",
        "Model",
        None,
        None,
        None,
        None,
        model_rep,
        None,
        "GRAPH_VIEW",
        None,
    )

    return {"model": model_rep, "body": body_sub_rep, "reference": ref_sub_rep}


def ifc_dir(f: ifcopenshell.file, vec: Tuple[float, float, float]):
    return f.create_entity("IfcDirection", to_real(vec))


def ensure_guid_consistency(a: Assembly, project_prefix):
    """Function to edit the global IDs of your elements when they are arbitrarily created from upstream data dump"""
    a.ifc_store.sync()
    ensure_uniqueness = dict()
    for p in a.get_all_parts_in_assembly():
        p.guid = create_guid(project_prefix + p.name)
        if p.guid in ensure_uniqueness.keys():
            # p_other = ensure_uniqueness.get(p.guid)
            # ancestors1 = p.get_ancestors()
            # ancestors2 = p_other.get_ancestors()
            raise ValueError(f"GUID Uniqueness not maintained for {p.name}")
        ensure_uniqueness[p.guid] = p
        for obj in p.get_all_physical_objects(sub_elements_only=True):
            obj.guid = create_guid(project_prefix + obj.name)

            if obj.guid in ensure_uniqueness.keys():
                conflicting_obj = ensure_uniqueness[obj.guid]
                # Handle special case BIM software outputs elements of same name
                # but one of them are opaque obstruction vols
                if conflicting_obj.opacity != 0.0:
                    conflicting_obj.name = conflicting_obj.name + "_INSU"
                    conflicting_obj.guid = create_guid(project_prefix + obj.name)
                    ensure_uniqueness[conflicting_obj.guid] = conflicting_obj
                elif obj.opacity != 0.0:
                    obj.name = obj.name + "_INSU"
                    obj.guid = create_guid(project_prefix + obj.name)
                else:
                    raise ValueError(f"GUID Uniqueness not maintained for '{obj}' in [{a.name}]")

            ensure_uniqueness[obj.guid] = obj


def create_guid(name=None):
    """Creates a guid from a random name or bytes or generates a random guid"""
    import hashlib
    import uuid

    if name is None:
        hexdig = uuid.uuid1().hex
    else:
        if type(name) != bytes:
            n = name.encode()
        else:
            n = name
        hexdig = hashlib.md5(n).hexdigest()
    result = ifcopenshell.guid.compress(hexdig)
    return result


def ifc_p(f: ifcopenshell.file, p):
    return f.create_entity("IfcCartesianPoint", to_real(p))


def create_ifc_placement(f: ifcopenshell.file, origin=ifco.O, loc_z=ifco.Z, loc_x=ifco.X):
    """
    Creates an IfcAxis2Placement3D from Location, Axis and RefDirection specified as Python tuples

    :param f:
    :param origin:
    :param loc_z:
    :param loc_x:
    :return:
    """

    ifc_loc_z = f.create_entity("IfcDirection", to_real(loc_z))
    ifc_loc_x = f.create_entity("IfcDirection", to_real(loc_x))
    return f.create_entity("IfcAxis2Placement3D", ifc_p(f, origin), ifc_loc_z, ifc_loc_x)


def create_local_placement(f: ifcopenshell.file, origin=ifco.O, loc_z=ifco.Z, loc_x=ifco.X, relative_to=None):
    """
    Creates an IfcLocalPlacement from Location, Axis and RefDirection,
    specified as Python tuples, and relative placement

    :param f:
    :param origin:
    :param loc_z:
    :param loc_x:
    :param relative_to:
    :return: IFC local placement
    """

    axis2placement = create_ifc_placement(f, origin, loc_z, loc_x)
    ifclocalplacement2 = f.create_entity(
        "IfcLocalPlacement", PlacementRelTo=relative_to, RelativePlacement=axis2placement
    )
    return ifclocalplacement2


def create_new_ifc_file(file_name, schema):
    import datetime

    from ..core.utils import get_version

    f = ifcopenshell.file(schema=schema)
    ada_ver = get_version()
    f.wrapped_data.header.file_name.name = file_name
    f.wrapped_data.header.file_name.time_stamp = (
        datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).astimezone().replace(microsecond=0).isoformat()
    )

    ver_str = f'IfcOpenShell: "{ifcopenshell.version}", ADA: "{ada_ver}"'

    f.wrapped_data.header.file_name.preprocessor_version = ver_str
    f.wrapped_data.header.file_name.originating_system = ver_str
    f.wrapped_data.header.file_name.authorization = "Nobody"
    length_unit = f.createIfcSIUnit(None, "LENGTHUNIT", None, "METRE")
    f.createIfcUnitAssignment((length_unit,))
    # f.wrapped_data.header.file_description.description = ("ViewDefinition[DesignTransferView]",)
    return f


def assembly_to_ifc_file(a: "Assembly"):
    return generate_tpl_ifc_file(a.name, a.metadata["project"], a.metadata["schema"], a.units, a.user)


def generate_tpl_ifc_file(file_name, project, schema, units, user):
    """

    :param file_name:
    :param project:
    :param schema:
    :param units:
    :param user:
    :type user: ada.config.User
    :return:
    """
    import time

    from ada.base.units import Units

    from .ifc_template import tpl_create

    timestamp = time.time()
    timestring = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(timestamp))
    application, application_version = "IfcOpenShell", "0.6"
    project_globalid = create_guid()
    if units == Units.M:
        units_str = "$,.METRE."
    elif units == Units.MM:
        units_str = ".MILLI.,.METRE."
    else:
        raise ValueError(f'Unrecognized unit prefix "{units}"')
    ifc_file = tpl_create(
        file_name + ".ifc",
        timestring,
        user.org_name,
        user.user_id,
        schema,
        application_version,
        int(timestamp),
        application,
        project_globalid,
        project,
        units_str,
        user.org_name,
    )

    return ifc_file


def create_ifcpolyline(ifcfile, point_list):
    """
    Creates an IfcPolyLine from a list of points, specified as Python tuples

    :param ifcfile:
    :param point_list:
    :return:
    """
    ifcpts = []
    for p_in in point_list:
        p = [float(x) for x in p_in]
        point = ifcfile.createIfcCartesianPoint(p)
        ifcpts.append(point)
    polyline = ifcfile.createIfcPolyLine(ifcpts)
    return polyline


def create_axis(f, points, context):
    polyline = create_ifcpolyline(f, points)
    return f.createIfcShapeRepresentation(context, "Axis", "Curve3D", [polyline])


def create_ifcindexpolyline(ifcfile, points, seg_index):
    """
    Assumes a point list whereas all points that are to be used for creating arc-segments will have 4 values
    (x,y,z,r) instead of 3 (x,y,z)

    #https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/link/annex-e.htm

    :param ifcfile:
    :param points:
    :param seg_index:
    :return:
    """
    ifc_segments = []
    for seg_ind in seg_index:
        if len(seg_ind) == 2:
            ifc_segments.append(ifcfile.createIfcLineIndex(seg_ind))
        elif len(seg_ind) == 3:
            ifc_segments.append(ifcfile.createIfcArcIndex(seg_ind))
        else:
            raise ValueError("Unrecognized number of values")

    # TODO: Investigate using 2DLists instead is it could reduce complexity?
    # ifc_point_list = ifcfile.createIfcCartesianPointList2D(points)

    ifc_point_list = ifcfile.createIfcCartesianPointList3D(points)
    segindex = ifcfile.createIfcIndexedPolyCurve(ifc_point_list, ifc_segments)
    return segindex


def create_ifcindexpolyline2d(ifcfile, points2d, seg_index):
    """
    Assumes a point list whereas all points that are to be used for creating arc-segments will have 4 values
    (x,y,z,r) instead of 3 (x,y,z)

    #https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/link/annex-e.htm

    :param ifcfile:
    :param points2d:
    :param seg_index:
    :return:
    """
    ifc_segments = []
    for seg_ind in seg_index:
        if len(seg_ind) == 2:
            ifc_segments.append(ifcfile.createIfcLineIndex(seg_ind))
        elif len(seg_ind) == 3:
            ifc_segments.append(ifcfile.createIfcArcIndex(seg_ind))
        else:
            raise ValueError("Unrecognized number of values")

    # TODO: Investigate using 2DLists instead is it could reduce complexity?
    # ifc_point_list = ifcfile.createIfcCartesianPointList2D(points)

    ifc_point_list = ifcfile.createIfcCartesianPointList2D(points2d)
    segindex = ifcfile.createIfcIndexedPolyCurve(ifc_point_list, ifc_segments)
    return segindex


def create_ifcrevolveareasolid(f, profile, ifcaxis2placement, origin, revolve_axis, revolve_angle):
    """Creates an IfcExtrudedAreaSolid from a list of points, specified as Python tuples"""
    ifcorigin = f.create_entity("IfcCartesianPoint", to_real(origin))
    ifcaxis1dir = f.create_entity(
        "IfcAxis1Placement", ifcorigin, f.create_entity("IfcDirection", to_real(revolve_axis))
    )

    return f.create_entity("IfcRevolvedAreaSolid", profile, ifcaxis2placement, ifcaxis1dir, revolve_angle)


def create_IfcFixedReferenceSweptAreaSolid(f, curve, profile, ifcaxis2placement, start_param, end_param, fixed_ref):
    """

    :param f:
    :param profile:
    :param ifcaxis2placement:
    :param curve:
    :param start_param:
    :param end_param:
    :param fixed_ref:
    :return:
    """
    return f.create_entity(
        "IfcFixedReferenceSweptAreaSolid", profile, ifcaxis2placement, curve, start_param, end_param, fixed_ref
    )


def create_ifcextrudedareasolid(ifc_file, profile, ifcaxis2placement, extrude_dir, extrusion):
    """
    Creates an IfcExtrudedAreaSolid from a list of points, specified as Python tuples

    :param profile:
    :param ifcaxis2placement:
    :param extrude_dir:
    :param extrusion:
    :return:
    """

    ifcdir = ifc_file.createIfcDirection(extrude_dir)
    ifcextrudedareasolid = ifc_file.create_entity("IfcExtrudedAreaSolid", profile, ifcaxis2placement, ifcdir, extrusion)
    return ifcextrudedareasolid


def create_ifcrightcylinder(ifc_file, ifcaxis2placement, height, radius):
    """

    :param ifc_file:
    :param ifcaxis2placement:
    :param height:
    :param radius:
    :return:
    """
    ifcextrudedareasolid = ifc_file.createIfcRightCircularCylinder(ifcaxis2placement, height, radius)
    return ifcextrudedareasolid


def ifc_value_map(f, value):
    value_map = {str: "IfcText", float: "IfcReal", int: "IfcInteger", bool: "IfcBoolean"}
    if type(value) in (np.float64,):
        value = float(value)
    ifc_type = value_map.get(type(value), None)
    if ifc_type is None:
        logger.warning(f'Unable to find suitable IFC type for "{type(value)}". Will convert it to string')
        return f.create_entity("IfcText", str(value))

    return f.create_entity(ifc_type, value)


def ifc_list_value(f, name: str, list_value: list, owner_history):
    list_values = []
    for x in list_value:
        if isinstance(x, (list, tuple, np.ndarray)):
            list_values.append(ifc_list_value(f, f"{name}_sub", x, owner_history))
        elif isinstance(x, dict):
            list_values.append(create_property_set(f"{name}_sub", f, x, owner_history))
        else:
            list_values.append(ifc_value_map(f, x))

    return f.create_entity("IfcPropertyListValue", Name=name, ListValues=list_values)


def create_property_set(name, ifc_file, metadata_props, owner_history):
    properties = []

    for key, value in metadata_props.items():
        if isinstance(value, (list, tuple, np.ndarray)):
            properties.append(ifc_list_value(ifc_file, key, value, owner_history))
        elif isinstance(value, dict):
            if len(value.keys()) == 0:
                continue
            properties.append(create_property_set(f"{name}_sub", ifc_file, value, owner_history))
        else:
            properties.append(
                ifc_file.create_entity(
                    "IfcPropertySingleValue",
                    Name=key,
                    NominalValue=ifc_value_map(ifc_file, value),
                )
            )

    atts = {
        "GlobalId": ifcopenshell.guid.new(),
        "OwnerHistory": owner_history,
        "Name": name,
        "HasProperties": properties,
    }

    return ifc_file.create_entity("IfcPropertySet", **atts)


def add_properties_to_elem(name, ifc_file, ifc_elem, elem_props, owner_history):
    logger.info(f'Adding "{name}" properties to IFC Element "{ifc_elem}"')

    props = create_property_set(name, ifc_file, elem_props, owner_history=owner_history)
    ifc_file.createIfcRelDefinesByProperties(
        create_guid(),
        owner_history,
        "Properties",
        None,
        [ifc_elem],
        props,
    )


def write_elem_property_sets(metadata_props, elem, f, owner_history) -> None:
    from ada.config import Settings

    if len(metadata_props.keys()) == 0 or Settings.model_export.export_props is False:
        return None

    if type(list(metadata_props.values())[0]) is dict:
        for pro_id, prop_ in metadata_props.items():
            add_properties_to_elem(pro_id, f, elem, prop_, owner_history=owner_history)
    else:
        add_properties_to_elem("Properties", f, elem, metadata_props, owner_history=owner_history)


def to_real(v) -> float | list[float]:
    from ada import Node

    if type(v) is float:
        return v
    elif type(v) is tuple:
        return [float(x) for x in v]
    elif type(v) is list:
        if type(v[0]) is float:
            return v
        else:
            return [float(x) for x in v]
    elif type(v) is Node:
        return v.p.astype(float).tolist()
    else:
        return v.astype(float).tolist()


def add_negative_extrusion(f, origin, loc_z, loc_x, depth, points, parent):
    context = f.by_type("IfcGeometricRepresentationContext")[0]
    owner_history = f.by_type("IfcOwnerHistory")[0]

    # Create and associate an opening for the window in the wall
    opening_placement = create_local_placement(f, origin, loc_z, loc_x, parent.ObjectPlacement)
    opening_axis_placement = create_ifc_placement(f, origin, loc_z, loc_x)
    polyline = create_ifcpolyline(f, points)
    ifcclosedprofile = f.createIfcArbitraryClosedProfileDef("AREA", None, polyline)

    opening_solid = create_ifcextrudedareasolid(f, ifcclosedprofile, opening_axis_placement, loc_z, depth)
    opening_representation = f.createIfcShapeRepresentation(context, "Body", "SweptSolid", [opening_solid])
    opening_shape = f.createIfcProductDefinitionShape(None, None, [opening_representation])
    opening_element = f.createIfcOpeningElement(
        create_guid(),
        owner_history,
        "Opening",
        "Door opening",
        None,
        opening_placement,
        opening_shape,
        None,
    )
    f.createIfcRelVoidsElement(create_guid(), owner_history, None, None, parent, opening_element)

    return opening_element


def add_colour(
    f,
    ifc_body: Union[List[ifcopenshell.entity_instance], ifcopenshell.entity_instance],
    name,
    colour,
    transparency=0.0,
    use_surface_style_rendering=False,
) -> None:
    """Add IFcSurfaceStyle using either IfcSurfaceStyleRendering or IfcSurfaceStyleShading"""
    if colour is None:
        return None
    colour = f.createIfcColourRgb(name, colour[0], colour[1], colour[2])

    if use_surface_style_rendering:
        surfaceStyleShading = f.createIFCSURFACESTYLERENDERING(colour, transparency)
    else:
        surfaceStyleShading = f.createIfcSurfaceStyleShading()
        surfaceStyleShading.SurfaceColour = colour

    surfaceStyle = f.createIfcSurfaceStyle(colour.Name, "BOTH", (surfaceStyleShading,))
    presStyleAssign = f.createIfcPresentationStyleAssignment((surfaceStyle,))
    if type(ifc_body) in [list, tuple]:
        for ifc_b in ifc_body:
            f.createIfcStyledItem(ifc_b, (presStyleAssign,), colour.Name)
    else:
        f.createIfcStyledItem(ifc_body, (presStyleAssign,), colour.Name)


def calculate_unit_scale(file):
    units = file.by_type("IfcUnitAssignment")[0]
    unit_scale = 1
    for unit in units.Units:
        if not hasattr(unit, "UnitType") or unit.UnitType != "LENGTHUNIT":
            continue
        while unit.is_a("IfcConversionBasedUnit"):
            unit_scale *= unit.ConversionFactor.ValueComponent.wrappedValue
            unit = unit.ConversionFactor.UnitComponent
        if unit.is_a("IfcSIUnit"):
            unit_scale *= get_prefix_multiplier(unit.Prefix)

    return unit_scale


def get_unit_type(file: ifcopenshell.file):
    from ada.base.units import Units

    value = calculate_unit_scale(file)
    if value == 0.001:
        return Units.MM
    elif value == 1:
        return Units.M
    else:
        raise NotImplementedError(f'Unit scale of "{value}" is not yet supported')


def scale_ifc_file_object(ifc_file, scale_factor):
    """
    Scale length factor to meter

    :param ifc_file:
    :return:
    """

    s = ifcopenshell.ifcopenshell_wrapper.schema_by_name("IFC4")
    classes_to_modify = {}
    for d in s.declarations():
        if not hasattr(d, "all_attributes") or "IfcLength" not in str(d.all_attributes()):
            continue
        attributes_to_modify = []
        for attribute in d.all_attributes():
            if "IfcLength" in str(attribute):
                attributes_to_modify.append(attribute.name())
        classes_to_modify[d.name()] = attributes_to_modify

    def scale_all(obj, sf):
        def serialize(obj_):
            """Recursively walk object's hierarchy."""
            if isinstance(obj_, (int, float)):
                return obj_ * sf
            elif isinstance(obj_, list):
                return [serialize(item) for item in obj_]
            elif isinstance(obj_, tuple):
                return tuple(serialize([item for item in obj_]))
            else:
                try:
                    if obj_.is_a("IfcLengthMeasure") is True:
                        obj_.wrappedValue = obj_.wrappedValue * sf
                        return obj_
                    elif obj_.is_a("IfcReal") is True:
                        obj_.wrappedValue = obj_.wrappedValue * sf
                        return obj_
                    elif obj_.is_a("IfcInteger") is True:
                        obj_.wrappedValue = int(obj_.wrappedValue * sf)
                        return obj_
                    elif obj_.is_a("IfcPlaneAngleMeasure") is True:
                        return obj_
                    elif obj_.is_a("IfcPressureMeasure") or obj_.is_a("IfcModulusOfElasticityMeasure"):
                        # sf is a length unit.
                        conv_unit = 1 / sf**2
                        obj_.wrappedValue = obj_.wrappedValue * conv_unit
                        return obj_
                    elif obj_.is_a("IfcMassDensityMeasure"):
                        conv_unit = 1 / sf**3
                        obj_.wrappedValue = obj_.wrappedValue * conv_unit
                        return obj_
                    # Unit-less
                    elif obj_.is_a("IfcText") is True or obj_.is_a("IfcPositiveRatioMeasure") is True:
                        return obj_
                    elif obj_.is_a("IfcThermalExpansionCoefficientMeasure") or obj_.is_a(
                        "IfcSpecificHeatCapacityMeasure"
                    ):
                        return obj_
                    elif obj_.is_a("IfcLogical") is True:
                        return obj_
                except Exception as er:
                    raise ValueError(f"Error {er}")

                raise ValueError(f'Unknown entity "{type(obj_)}", "{obj_}"')

        return serialize(obj)

    for ifc_class, attributes in classes_to_modify.items():
        for element in ifc_file.by_type(ifc_class):
            for attribute in attributes:
                old_val = getattr(element, attribute)
                if old_val is None:
                    continue
                # setattr(element, attribute, scale_all(old_val, scale_factor))
                # new_val = getattr(element, attribute)
    return ifc_file


def merge_existing(original_file, source_file, new_file):
    source = ifcopenshell.open(source_file)
    f = ifcopenshell.open(original_file)
    original_project = f.by_type("IfcProject")[0]
    merged_project = f.add(source.by_type("IfcProject")[0])
    for element in source.by_type("IfcRoot"):
        f.add(element)
    for inverse in f.get_inverse(merged_project):
        ifcopenshell.util.element.replace_attribute(inverse, merged_project, original_project)
    f.remove(merged_project)
    f.write(str(new_file))


def patch(f, source_file):
    """

    :param f:
    :param source_file:
    :return:
    """
    source = ifcopenshell.open(source_file)
    original_project = f.by_type("IfcProject")[0]
    merged_project = f.add(source.by_type("IfcProject")[0])
    for element in source.by_type("IfcRoot"):
        f.add(element)
    for inverse in f.get_inverse(merged_project):
        ifcopenshell.util.element.replace_attribute(inverse, merged_project, original_project)
    f.remove(merged_project)


def merge_ifc_files(parent_dir, output_file_name, clean_files=False, include_elements=None, exclude_elements=None):
    """

    :param parent_dir:
    :param output_file_name:
    :param clean_files:
    :param include_elements:
    :param exclude_elements:
    :return:
    """
    import pathlib
    import time

    parent_dir = pathlib.Path(parent_dir)
    files = get_list_of_files(parent_dir, ".ifc")
    f = ifcopenshell.open(files[0])
    for i, fp in enumerate(files[1:]):
        checkpoint = time.time()
        print(f'merging products ({i + 1} of {len(files) - 1}) from "{fp}"')
        fn = ifcopenshell.open(fp)
        print(f"file opened in {time.time() - checkpoint:.2f} seconds")
        checkpoint = time.time()
        for product in fn.by_type("IfcProduct"):
            f.add(product)
        print(f"Products added in {time.time() - checkpoint:.2f} seconds")
        del fn
    out_file_name = str((parent_dir / output_file_name).with_suffix(".ifc"))
    print(f'Writing file "{out_file_name}"')
    checkpoint = time.time()
    f.write(out_file_name)
    print(f"File written in {time.time() - checkpoint:.2f} seconds")


def convert_bm_jusl_to_ifc(bm: "Beam") -> int:
    """
    IfcCardinalPointReference


    1.      bottom left
    2.      bottom centre
    3.      bottom right
    4.      mid-depth left
    5.      mid-depth centre
    6.      mid-depth right
    7.      top left
    8.      top centre
    9.      top right
    10.     geometric centroid
    11.     bottom in line with the geometric centroid
    12.     left in line with the geometric centroid
    13.     right in line with the geometric centroid
    14.     top in line with the geometric centroid
    15.     shear centre
    16.     bottom in line with the shear centre
    17.     left in line with the shear centre
    18.     right in line with the shear centre
    19.     top in line with the shear centre

    https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/schema/ifcmaterialresource/lexical/ifccardinalpointreference.htm
    """
    jusl = bm.jusl
    jt = bm.JUSL_TYPES

    jusl_map = {jt.NA: 5, jt.TOS: 8}

    jusl_val = jusl_map.get(jusl, None)

    if jusl_val is None:
        if jusl != jt.NA:
            logger.error(f'Unknown JUSL value "{jusl}". Using NA')
        return 5

    return jusl_val


def scale_ifc_file(current_ifc, new_ifc):
    oval = calculate_unit_scale(current_ifc)
    nval = calculate_unit_scale(new_ifc)
    if oval != nval:
        logger.error("Running Unit Conversion on IFC import. This is still highly unstable")
        # length_unit = f.createIfcSIUnit(None, "LENGTHUNIT", None, "METRE")
        # unit_assignment = f.createIfcUnitAssignment((length_unit,))
        new_file = scale_ifc_file_object(new_ifc, nval)
        return new_file


def tesselate_shape(shape, schema, tol):
    occ_string = ifcopenshell.geom.occ_utils.serialize_shape(shape)
    serialized_geom = ifcopenshell.geom.serialise(schema, occ_string)

    if serialized_geom is None:
        logger.debug("Starting serialization of geometry")
        serialized_geom = ifcopenshell.geom.tesselate(schema, occ_string, tol)

    return serialized_geom


def default_settings():
    ifc_settings = ifcopenshell.geom.settings()
    ifc_settings.set(ifc_settings.USE_PYTHON_OPENCASCADE, True)
    ifc_settings.set(ifc_settings.SEW_SHELLS, True)
    ifc_settings.set(ifc_settings.WELD_VERTICES, True)
    ifc_settings.set(ifc_settings.INCLUDE_CURVES, True)
    ifc_settings.set(ifc_settings.USE_WORLD_COORDS, True)
    ifc_settings.set(ifc_settings.VALIDATE_QUANTITIES, True)
    return ifc_settings


def export_transform(f: ifcopenshell.file, transform: Transform):
    from ada.core.constants import X

    # https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/annex/annex-e/mapped-shape-with-multiple-items.ifc
    # axis1 = f.create_entity("")
    f.create_entity(
        "IfcCartesianTransformationOperator",
        ifc_dir(f, X),
    )
    raise NotImplementedError()


def get_all_subtypes(entity: ifcopenshell.ifcopenshell_wrapper.entity, subtypes=None):
    subtypes = [] if subtypes is None else subtypes
    for subtype in entity.subtypes():
        if subtype.is_abstract() is False:
            subtypes.append(subtype.name())
        get_all_subtypes(subtype, subtypes)
    return subtypes


def get_representation_items(f: ifcopenshell.file, ifc_elem: ifcopenshell.entity_instance):
    geom_items = [
        "IfcTriangulatedFaceSet",
        "IfcExtrudedAreaSolid",
        "IfcExtrudedAreaSolid",
        "IfcExtrudedAreaSolidTapered",
        "IfcRevolvedAreaSolid",
        "IfcClosedShell",
    ]
    geom_lower = [i.lower() for i in geom_items]
    return list(
        filter(
            lambda x: hasattr(x, "StyledByItem") and x.is_a().lower() in geom_lower,
            f.traverse(ifc_elem),
        )
    )


def ifc_file_to_stp(ifc_file, stp_file, step_schema="AP242", step_assembly_mode=1):
    """Convert an IFC file to STEP file"""

    import os
    import pathlib

    from OCC.Core.IFSelect import IFSelect_RetError
    from OCC.Core.Interface import Interface_Static_SetCVal
    from OCC.Core.STEPConstruct import stepconstruct_FindEntity
    from OCC.Core.STEPControl import STEPControl_AsIs, STEPControl_Writer
    from OCC.Core.TCollection import TCollection_HAsciiString

    settings = ifcopenshell.geom.settings()
    settings.set(settings.USE_PYTHON_OPENCASCADE, True)
    # settings.set(settings.SEW_SHELLS, False)
    # settings.set(settings.WELD_VERTICES, True)
    # settings.set(settings.INCLUDE_CURVES, False)
    settings.set(settings.USE_WORLD_COORDS, True)
    # settings.set(settings.VALIDATE_QUANTITIES, False)

    f = ifcopenshell.open(ifc_file)
    writer = STEPControl_Writer()
    fp = writer.WS().TransferWriter().FinderProcess()

    Interface_Static_SetCVal("write.step.schema", step_schema)
    Interface_Static_SetCVal("write.precision.mode", "1")
    Interface_Static_SetCVal("write.step.assembly", str(step_assembly_mode))
    iterator = ifcopenshell.geom.iterator(settings, f)
    iterator.initialize()

    while True:
        shape = iterator.get()
        if shape:
            geom = shape.geometry
            name = f.by_id(shape[0].id).Name
            Interface_Static_SetCVal("write.step.product.name", name)
            writer.Transfer(geom, STEPControl_AsIs)
            item = stepconstruct_FindEntity(fp, geom)
            if not item:
                logger.debug("STEP item not found for FindEntity")
            else:
                item.SetName(TCollection_HAsciiString(name))
        if not iterator.next():
            break

    destination_file = pathlib.Path(stp_file).with_suffix(".stp")
    os.makedirs(destination_file.parent, exist_ok=True)

    status = writer.Write(str(destination_file))
    if int(status) > int(IFSelect_RetError):
        raise Exception("Error during write operation")

    print(f'step file created at "{destination_file}"')



from typing import List

from ada import FEM
from ada.core.utils import Counter

from .concepts import GmshSession, GmshTask


def multisession_gmsh_tasker(fem: FEM, gmsh_tasks: List[GmshTask]):
    """Run multiple meshing operations within a single GmshSession."""

    model_names = Counter(1, "gmsh")
    with GmshSession(silent=True) as gs:
        for gtask in gmsh_tasks:
            gs.model.add(next(model_names))
            gs.options = gtask.options
            for obj in gtask.ada_obj:
                gs.add_obj(obj, gtask.geom_repr)
            gs.mesh(gtask.mesh_size)

            # TODO: Add operand type += for FEM
            tmp_fem = gs.get_fem()
            tmp_fem.parent = fem.parent
            fem += tmp_fem
            gs.model_map = dict()
    return fem



from __future__ import annotations

import json
import os
import pathlib
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ada.config import get_logger

if TYPE_CHECKING:
    from ada import Assembly

logger = get_logger()


@dataclass
class CacheStore:
    name: str
    state_file: pathlib.Path = field(default=None)
    cache_file: pathlib.Path = field(default=None)

    _cache_loaded: bool = False

    def __post_init__(self):
        state_path = pathlib.Path("").parent.resolve().absolute() / ".state" / self.name
        self.state_file = state_path.with_suffix(".json")
        self.cache_file = state_path.with_suffix(".h5")

    def sync(self, assembly: Assembly, clear_cache=False):
        if self.cache_file.exists() and clear_cache:
            os.remove(self.cache_file)
        if self.state_file.exists() and clear_cache:
            os.remove(self.state_file)

        self._cache_loaded = False
        self.from_cache(assembly)

    def from_cache(self, assembly: Assembly, input_file=None):
        is_cache_outdated = self.is_cache_outdated(input_file)
        if input_file is None and is_cache_outdated is False:
            self.read_cache(assembly)
            return True

        if is_cache_outdated is False and self._cache_loaded is False:
            self.read_cache(assembly)
            return True
        elif is_cache_outdated is False and self._cache_loaded is True:
            return True
        else:
            return False

    def _get_file_state(self):
        state_file = self.state_file
        if state_file.exists() is True:
            with open(state_file, "r") as f:
                state = json.load(f)
                return state
        return dict()

    def _update_file_state(self, input_file=None):
        in_file = pathlib.Path(input_file)
        fna = in_file.name
        last_modified = os.path.getmtime(in_file)
        state_file = self.state_file
        state = self._get_file_state()

        state.get(fna, dict())
        state[fna] = dict(lm=last_modified, fp=str(in_file))

        os.makedirs(state_file.parent, exist_ok=True)

        with open(state_file, "w") as f:
            json.dump(state, f, indent=4)

    def to_cache(self, assembly: Assembly, input_file, write_to_cache: bool):
        self._update_file_state(input_file)
        if write_to_cache:
            self.update_cache(assembly)

    def read_cache(self, assembly: Assembly):
        from ada.cache.reader import read_assembly_from_cache

        read_assembly_from_cache(self.cache_file, assembly)
        print(f"Finished Loading model from cache {self.cache_file}")

    def update_cache(self, assembly: Assembly):
        from ada.cache.writer import write_assembly_to_cache

        write_assembly_to_cache(assembly, self.cache_file)

    def is_cache_outdated(self, input_file=None):
        is_cache_outdated = False
        state = self._get_file_state()

        for name, props in state.items():
            in_file = pathlib.Path(props.get("fp"))
            last_modified_state = props.get("lm")
            if in_file.exists() is False:
                is_cache_outdated = True
                break

            last_modified = os.path.getmtime(in_file)
            if last_modified != last_modified_state:
                is_cache_outdated = True
                break

        if self.cache_file.exists() is False:
            logger.debug("Cache file not found")
            is_cache_outdated = True

        if input_file is not None:
            curr_in_file = pathlib.Path(input_file)
            if curr_in_file.name not in state.keys():
                is_cache_outdated = True

        return is_cache_outdated



import re
from typing import TYPE_CHECKING

from ada import Material
from ada.core.utils import roundoff
from ada.materials.metals import CarbonSteel, PlasticityModel

from .helper_utils import _re_in

if TYPE_CHECKING:
    from ada import Assembly


def get_materials_from_bulk(assembly: "Assembly", bulk_str):
    re_str = (
        r"(\*Material,\s*name=.*?)(?=\*|\Z)(?!\*Elastic|\*Density|\*Plastic|"
        r"\*Damage Initiation|\*Damage Evolution|\*Expansion)"
    )
    re_materials = re.compile(re_str, _re_in)
    for m in re_materials.finditer(bulk_str):
        mat = mat_str_to_mat_obj(m.group())
        assembly.add_material(mat)


def mat_str_to_mat_obj(mat_str) -> Material:
    rd = roundoff

    # Name
    name = re.search(r"name=(.*?)\n", mat_str, _re_in).group(1).split("=")[-1].strip()

    # Density
    density_ = re.search(r"\*Density\n(.*?)(?:,|$)", mat_str, _re_in)
    if density_ is not None:
        density = rd(density_.group(1).strip().split(",")[0].strip(), 10)
    else:
        print('No density flag found for material "{}"'.format(name))
        density = None

    # Elastic
    re_elastic_ = re.search(r"\*Elastic(?:,\s*type=(.*?)|)\n(.*?)(?:\*|$)", mat_str, _re_in)
    if re_elastic_ is not None:
        re_elastic = re_elastic_.group(2).strip().split(",")
        young, poisson = rd(re_elastic[0]), rd(re_elastic[1])
    else:
        print('No Elastic properties found for material "{name}"'.format(name=name))
        young, poisson = None, None

    # Plastic
    re_plastic_ = re.search(r"\*Plastic\n(.*?)(?:\*|\Z)", mat_str, _re_in)
    if re_plastic_ is not None:
        re_plastic = [tuple(x.split(",")) for x in re_plastic_.group(1).strip().splitlines()]
        sig_p = [rd(x[0]) for x in re_plastic]
        eps_p = [rd(x[1]) for x in re_plastic]
    else:
        eps_p, sig_p = None, None

    # Expansion
    re_zeta = re.search(r"\*Expansion(?:,\s*type=(.*?)|)\n(.*?)(?:\*|$)", mat_str, _re_in)
    if re_zeta is not None:
        zeta = float(re_zeta.group(2).split(",")[0].strip())
    else:
        zeta = 0.0

    # Return material object
    model = CarbonSteel(
        rho=density,
        E=young,
        v=poisson,
        zeta=zeta,
        plasticity_model=PlasticityModel(eps_p=eps_p, sig_p=sig_p),
    )
    return Material(name=name, mat_model=model)



from __future__ import annotations

from enum import Enum

import numpy as np

from ada.base.types import GeomRepr
from ada.config import get_logger

# The element names are based on the naming scheme by meshio

logger = get_logger()


class UnsupportedFeaShapeException(Exception):
    pass


class BaseShapeEnum(Enum):
    @classmethod
    def from_str(cls, value: str, default=None):
        if isinstance(value, cls):
            return value
        key_map = {x.value.lower(): x for x in cls}
        result = key_map.get(value.lower(), default)
        if result is None:
            raise UnsupportedFeaShapeException("Unsupported")

        return result

    @classmethod
    def get_all(cls) -> list:
        return [x for x in cls]

    def __gt__(self, other):
        return self.value > other.value


class LineShapes(BaseShapeEnum):
    LINE = "LINE"
    LINE3 = "LINE3"


class ShellShapes(BaseShapeEnum):
    TRI = "TRIANGLE"
    TRI6 = "TRIANGLE6"
    TRI7 = "TRIANGLE7"
    QUAD = "QUAD"
    QUAD8 = "QUAD8"
    QUAD9 = "QUAD9"


class SolidShapes(BaseShapeEnum):
    HEX8 = "HEXAHEDRON"
    HEX20 = "HEXAHEDRON20"
    HEX27 = "HEXAHEDRON27"
    TETRA = "TETRA"
    TETRA10 = "TETRA10"
    PYRAMID5 = "PYRAMID5"
    PYRAMID13 = "PYRAMID13"
    WEDGE = "WEDGE"
    WEDGE15 = "WEDGE15"


class ConnectorTypes(BaseShapeEnum):
    CONNECTOR = "CONNECTOR"


class MassTypes(BaseShapeEnum):
    MASS = "MASS"
    ROTARYI = "ROTARYI"


class SpringTypes(BaseShapeEnum):
    SPRING1 = "SPRING1"
    SPRING2 = "SPRING2"


class ShapeResolver:
    NUM_MAP = {
        1: MassTypes.get_all() + SpringTypes.get_all(),
        2: [LineShapes.LINE] + ConnectorTypes.get_all(),
        3: [LineShapes.LINE3, ShellShapes.TRI],
        4: [ShellShapes.QUAD, SolidShapes.TETRA],
        5: [SolidShapes.PYRAMID5],
        6: [ShellShapes.TRI6, SolidShapes.WEDGE],
        8: [SolidShapes.HEX8, ShellShapes.QUAD8],
        10: [SolidShapes.TETRA10],
        15: [SolidShapes.WEDGE15],
        20: [SolidShapes.HEX20],
        27: [SolidShapes.HEX27],
    }

    @staticmethod
    def get_el_type_from_str(el_type: str) -> LineShapes | ShellShapes | SolidShapes | None:
        for shape in [LineShapes, ShellShapes, SolidShapes, SpringTypes, ConnectorTypes]:
            try:
                result = shape.from_str(el_type)
            except UnsupportedFeaShapeException:
                continue

            return result

        return None

    @staticmethod
    def get_el_nodes_from_type(el_type: LineShapes | ShellShapes | SolidShapes):
        for num, el_types in ShapeResolver.NUM_MAP.items():
            if el_type in el_types:
                return num

        raise ValueError(f'element type "{el_type}" is not yet supported')

    @staticmethod
    def to_geom_repr(el_type):
        if isinstance(el_type, SolidShapes):
            return ElemType.SOLID
        elif isinstance(el_type, ShellShapes):
            return ElemType.SHELL
        elif isinstance(el_type, LineShapes):
            return ElemType.LINE
        else:
            raise ValueError(f'Unrecognized Shape Type: "{el_type}"')


class ElemType:
    SHELL = GeomRepr.SHELL
    SOLID = GeomRepr.SOLID
    LINE = GeomRepr.LINE

    LINE_SHAPES = LineShapes
    SHELL_SHAPES = ShellShapes
    SOLID_SHAPES = SolidShapes

    MASS_SHAPES = MassTypes
    CONNECTOR_SHAPES = ConnectorTypes

    all = [SHELL, SOLID, LINE]


class ElemShapeTypes:
    shell = ShellShapes
    solids = SolidShapes
    lines = LineShapes

    spring1n = ["SPRING1"]
    spring2n = ["SPRING2"]
    springs = spring1n + spring2n
    masses = ["MASS", "ROTARYI"]
    connectors = ["CONNECTOR", "CONN3D2"]
    other2n = connectors
    other = other2n


class ElemShape:
    TYPES = ElemShapeTypes

    def __init__(self, el_type, nodes):
        self.type = None
        self.nodes = None
        self._edges = None
        self._faces = None
        self.update(el_type, nodes)

    @property
    def edges(self):
        edges_seq = self.edges_seq
        if edges_seq is None:
            raise ValueError(f"Element type {self.type} is currently not supported for Visualization")
        if self._edges is None:
            self._edges = [self.nodes[e] for ed_seq in edges_seq for e in ed_seq]

        return self._edges

    def get_faces(self):
        from itertools import chain

        def hex_face_to_tris(q):
            return [(q[0], q[1], q[2]), (q[0], q[2], q[3])]

        if isinstance(self.type, SolidShapes):
            faces_seq = self.solids_face_seq
            if self.type in (SolidShapes.HEX8, SolidShapes.HEX20):
                faces_seq = list(chain.from_iterable([hex_face_to_tris(x) for x in faces_seq]))
        else:
            faces_seq = self.faces_seq

        if self._faces is None:
            self._faces = [self.nodes[e] for ed_seq in faces_seq for e in ed_seq]

        return self._faces

    @property
    def faces(self):
        if isinstance(self.type, SolidShapes):
            faces_seq = self.solids_face_seq
        else:
            faces_seq = self.faces_seq

        if faces_seq is None:
            raise ValueError(f"Element type {self.type} is currently not supported for Visualization")

        if self._faces is None:
            self._faces = [self.nodes[e] for ed_seq in faces_seq for e in ed_seq]

        return self._faces

    @property
    def elem_type_group(self):
        if isinstance(self.type, SolidShapes):
            return ElemType.SOLID
        elif isinstance(self.type, ShellShapes):
            return ElemType.SHELL
        elif isinstance(self.type, LineShapes):
            return ElemType.LINE
        elif isinstance(self.type, SpringTypes):
            return ElemType.LINE
        else:
            raise ValueError(f'Unrecognized Element Type: "{self.type}"')

    def update(self, el_type=None, nodes=None):
        if el_type is not None:
            if isinstance(el_type, str):
                el_type = ShapeResolver.get_el_type_from_str(el_type)
            if el_type is None:
                raise ValueError(f'Currently unsupported element type "{el_type}".')
            self.type = el_type

        nodes = self.nodes if nodes is None else nodes
        num_nodes = ShapeResolver.get_el_nodes_from_type(self.type)
        if len(nodes) != num_nodes:
            raise ValueError(f'Number of passed nodes "{len(nodes)}" does not match expected "{num_nodes}" ')

        self.nodes = nodes
        self._edges = None

    @property
    def edges_seq(self) -> np.ndarray | None:
        from .lines import line_edges
        from .shells import shell_edges
        from .solids import solid_edges

        edge_map = {
            ElemType.LINE: line_edges,
            ElemType.SHELL: shell_edges,
            ElemType.SOLID: solid_edges,
            SpringTypes.SPRING2: line_edges,
        }
        generalized_type = self.type
        edges_repo = edge_map[self.elem_type_group]
        if generalized_type not in edges_repo.keys():
            logger.error(f"Element type {self.type} is currently not supported")
            return None

        return edges_repo[generalized_type]

    @property
    def faces_seq(self):
        from .shells import shell_faces
        from .solids import solid_faces

        face_map = {ElemType.LINE: None, ElemType.SHELL: shell_faces, ElemType.SOLID: solid_faces}
        generalized_type = self.type
        faces_repo = face_map[self.elem_type_group]
        if generalized_type not in faces_repo.keys():
            raise ValueError(f"Element type {self.type} is currently not supported for Visualization")

        return faces_repo[generalized_type]

    @property
    def spring_edges(self):
        if self.type not in ElemShapeTypes.springs:
            return None
        springs = dict(SPRING2=[[0, 1]])
        return springs[self.type]

    @property
    def solids_face_seq(self):
        from .solids import solid_faces

        solid_face_res = solid_faces.get(self.type, None)
        if solid_face_res is None:
            logger.error(f"Element type {self.type} is currently not supported")
            return None

        return solid_face_res

    @staticmethod
    def is_valid_elem(elem_type):
        valid_element_types = (
            LineShapes.get_all()
            + ShellShapes.get_all()
            + SolidShapes.get_all()
            + MassTypes.get_all()
            + SpringTypes.get_all()
            + ConnectorTypes.get_all()
        )
        valid_element_types_upper = [x.value.upper() for x in valid_element_types]
        value = elem_type.upper()
        if value in valid_element_types_upper:
            return True
        else:
            return False

    def __repr__(self):
        return f'{self.__class__.__name__}(Type: {self.type}, NodeIds: "{self.nodes}")'


def get_elem_type_group(el_type):
    if isinstance(el_type, SolidShapes):
        return ElemType.SOLID
    elif el_type in up(ShellShapes.get_all):
        return ElemType.SHELL
    elif el_type in up(LineShapes.get_all):
        return ElemType.LINE
    else:
        raise ValueError(f'Unrecognized Element Type: "{el_type}"')


def up(variables):
    return [v.upper() for v in variables]



import logging
import os
import pathlib
from dataclasses import dataclass


@dataclass
class ModelExportOptions:
    export_props: bool = True
    import_props: bool = True
    include_ecc = True


def _get_platform_home():
    """Home location for each platform"""
    # _platform_home = dict(win32="C:/ADA", linux="/home/ADA", linux2="/home/ADA", macos="/home/ADA")
    # return _platform_home[sys.platform]

    return pathlib.Path.home() / "ADA"


class Settings:
    """The Properties object contains all general purpose properties relevant for Parts and Assemblies"""

    point_tol = 1e-4
    precision = 6
    mtol = 1e-3
    mmtol = 1
    valid_units = ["m", "mm"]

    safe_deletion = True

    convert_bad_names = False
    convert_bad_names_for_fem = True
    use_occ_bounding_box_algo = False

    force_param_profiles = True
    silence_display = False
    use_experimental_cache = False

    # IFC export settings
    model_export: ModelExportOptions = ModelExportOptions()

    # FEM analysis settings
    if os.getenv("ADA_execute_dir", None) is not None:
        execute_dir = pathlib.Path(os.getenv("ADA_execute_dir", None))
    else:
        execute_dir = None

    # Visualization Settings
    use_new_visualize_api = False

    # Code Aster conversion specific settings
    ca_experimental_id_numbering = False

    debug = False
    _home = _get_platform_home()
    scratch_dir = pathlib.Path(os.getenv("ADA_scratch_dir", f"{_home}/Analyses"))
    temp_dir = pathlib.Path(os.getenv("ADA_temp_dir", f"{_home}/temp"))
    debug_dir = pathlib.Path(os.getenv("ADA_log_dir", f"{_home}/logs"))
    test_dir = pathlib.Path(os.getenv("ADA_test_dir", f"{_home}/tests"))
    tools_dir = pathlib.Path(os.getenv("ADA_tools_dir", f"{_home}/tools"))

    fem_exe_paths = dict(abaqus=None, ccx=None, sestra=None, usfos=None, code_aster=None)

    use_duplicate_log_filter = True

    @classmethod
    def default_ifc_settings(cls):
        from ada.ifc.utils import default_settings

        return default_settings()


class DuplicateFilter(logging.Filter):
    MAX_NUM = 3

    def __init__(self, name="", logger=None):
        super().__init__(name)
        self.last_log = None
        self.count = 0
        self.logger = logger

    def filter(self, record):
        # add other fields if you need more granular comparison, depends on your app
        if getattr(record, "suppress_filters", False):
            return True

        max_num = self.MAX_NUM
        current_log = (record.module, record.levelno, record.msg)

        if current_log == self.last_log:
            self.count += 1
            if self.count == max_num:
                record.msg = f"The previous message is repeated {self.count} times and will be ignored."
                return True
            elif self.count > max_num:
                return False

            return True

        if self.count > max_num:
            sup_str = f"It was suppressed {self.count - max_num} time(s)."
            self.logger.info(
                f"... The previous log message was suppressed after {max_num} repetitions. {sup_str}",
                extra={"suppress_filters": True},
            )

        self.last_log = current_log
        self.count = 1
        return True


def get_logger():
    _logger = logging.getLogger("ada")
    if Settings.use_duplicate_log_filter:
        _logger.addFilter(DuplicateFilter(logger=_logger))
    return _logger



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.config import get_logger

from .exceptions import NoIfcAxesAttachedError, UnableToConvertBoolResToBeamException
from .read_beams import import_ifc_beam
from .read_pipe import import_pipe_segment
from .read_plates import import_ifc_plate
from .read_shapes import import_ifc_shape

if TYPE_CHECKING:
    from ada.ifc.store import IfcStore

logger = get_logger()


def import_physical_ifc_elem(product, name, ifc_store: IfcStore):
    pr_type = product.is_a()

    if pr_type in ["IfcBeamStandardCase", "IfcBeam"]:
        try:
            return import_ifc_beam(product, name, ifc_store)
        except (NoIfcAxesAttachedError, UnableToConvertBoolResToBeamException) as e:
            logger.debug(e)
            pass
    if pr_type in ["IfcPlateStandardCase", "IfcPlate"]:
        try:
            return import_ifc_plate(product, name, ifc_store)
        except NoIfcAxesAttachedError as e:
            logger.debug(e)
            pass

    if product.is_a("IfcOpeningElement") is True:
        logger.info(f'skipping opening element "{product}"')
        return None

    if product.is_a() in ("IfcPipeSegment", "IfcPipeFitting"):
        return import_pipe_segment(product, name, ifc_store)

    if product.is_a("IfcPipeFitting"):
        logger.info('"IfcPipeFitting" is not yet added')

    obj = import_ifc_shape(product, name, ifc_store)

    return obj



import os

from ada.concepts.spatial import Assembly, Part

from .read_constraints import get_bcs, get_constraints
from .read_elements import get_elements, get_mass, get_springs
from .read_materials import get_materials
from .read_nodes import get_nodes, renumber_nodes
from .read_sections import get_sections
from .read_sets import get_sets


def read_fem(fem_file: os.PathLike, fem_name: str = None):
    """Import contents from a Sesam fem file into an assembly object"""

    print("Starting import of Sesam input file")
    part_name = "T1" if fem_name is None else fem_name
    with open(fem_file, "r") as d:
        part = read_sesam_fem(d.read(), part_name)

    return Assembly("TempAssembly") / part


def read_sesam_fem(bulk_str, part_name) -> Part:
    """Reads the content string of a Sesam input file and converts it to FEM objects"""

    part = Part(part_name)
    fem = part.fem

    fem.nodes = get_nodes(bulk_str, fem)
    elements, mass_elem, spring_elem, el_id_map = get_elements(bulk_str, fem)
    fem.elements = elements
    fem.elements.build_sets()
    part._materials = get_materials(bulk_str, part)
    fem.sections = get_sections(bulk_str, fem, mass_elem, spring_elem)
    fem.elements += get_mass(bulk_str, part.fem, mass_elem)
    fem.springs = get_springs(bulk_str, fem, spring_elem)
    fem.sets = part.fem.sets + get_sets(bulk_str, fem)
    fem.constraints.update(get_constraints(bulk_str, fem))
    fem.bcs += get_bcs(bulk_str, fem)
    renumber_nodes(bulk_str, fem)
    fem.elements.renumber(renumber_map=el_id_map)

    print(8 * "-" + f'Imported "{fem.instance_name}"')
    return part



import xml.etree.ElementTree as ET
from itertools import chain
from typing import List, Tuple

import numpy as np

from ada import Beam, Node, Part
from ada.concepts.containers import Beams
from ada.config import get_logger
from ada.core.exceptions import VectorNormalizeError

logger = get_logger()


def get_beams(xml_root: ET.Element, parent: Part) -> Beams:
    all_beams = xml_root.findall(".//straight_beam") + xml_root.findall(".//curved_beam")
    return Beams(chain.from_iterable([el_to_beam(bm_el, parent) for bm_el in all_beams]), parent)


def el_to_beam(bm_el: ET.Element, parent: Part) -> List[Beam]:
    name = bm_el.attrib["name"]
    xv, yv, zv = get_curve_orientation(bm_el)
    segs = []
    prev_bm = None
    for seg in bm_el.iterfind(".//straight_segment"):
        cur_bm = seg_to_beam(name, seg, parent, prev_bm, zv)
        if cur_bm is None:
            continue

        prev_bm = cur_bm
        segs += [cur_bm]

    for seg in bm_el.iterfind(".//curved_segment"):
        cur_bm = seg_to_beam(name, seg, parent, prev_bm, zv)
        if cur_bm is None:
            continue

        prev_bm = cur_bm
        segs += [cur_bm]

    if len(segs) > 0:
        apply_offsets_and_alignments(name, bm_el, segs)
    else:
        logger.warning(f"No segments found for beam {name}")

    return segs


def apply_offsets_and_alignments(name: str, bm_el: ET.Element, segs: list[Beam]):
    e1, e2, use_local = get_offsets(bm_el)
    alignment = get_alignment(bm_el, segs)

    if len(segs) > 1 and (e1 is not None or e2 is not None):
        logger.debug(f"Offset at end 1 for beam {name} is ignored as there are more than 1 segments")

    if e1 is not None:
        if use_local:
            e1_global = convert_offset_to_global_csys(e1, segs[0])
        else:
            e1_global = e1
        segs[0].e1 = e1_global
    if e2 is not None:
        if use_local:
            e2_global = convert_offset_to_global_csys(e2, segs[-1])
        else:
            e2_global = e2
        segs[-1].e2 = e2_global

    if alignment is not None:
        for seg in segs:
            seg.e1 = alignment if seg.e1 is None else seg.e1 + alignment
            seg.e2 = alignment if seg.e2 is None else seg.e2 + alignment


def get_offsets(bm_el: ET.Element) -> tuple[np.ndarray, np.ndarray, bool]:
    linear_offset = bm_el.find("curve_offset/linear_varying_curve_offset")

    end1_o = None
    end2_o = None
    use_local = False
    if linear_offset is not None:
        end1 = linear_offset.find("offset_end1")
        end2 = linear_offset.find("offset_end2")
        use_local = False if linear_offset.attrib["use_local_system"] == "false" else True
    else:
        end1 = bm_el.find(".//offset_end1")
        end2 = bm_el.find(".//offset_end2")

    if end1 is not None:
        end1_o = np.array(xyz_to_floats(end1))
    if end2 is not None:
        end2_o = np.array(xyz_to_floats(end2))

    return end1_o, end2_o, use_local


def get_alignment(bm_el: ET.Element, segments: list[Beam]):
    seg0 = segments[0]
    sec0 = seg0.section
    zv = seg0.up
    aligned_offset = bm_el.find("curve_offset/aligned_curve_offset")

    if aligned_offset is None:
        # if sec0.type == sec0.TYPES.ANGULAR:
        #     offset = zv * sec0.h / 2
        #     return offset
        return None

    alignment = aligned_offset.attrib.get("alignment")
    aligned_offset.attrib.get("constant_value")
    if alignment == "flush_top":
        if sec0.type == sec0.TYPES.ANGULAR:
            pass  # Angular profiles are already flush
        elif sec0.type == sec0.TYPES.TUBULAR:
            offset = -zv * sec0.r
            return offset
        else:
            offset = -zv * sec0.h / 2
            return offset


def convert_offset_to_global_csys(o: np.ndarray, bm: Beam):
    xv = bm.xvec
    yv = bm.yvec
    zv = bm.up
    return xv * o[0] + yv * o[1] + zv * o[2]


def apply_offset(o: np.ndarray, n: Node, bm: Beam):
    res = convert_offset_to_global_csys(o, bm)
    return n.p + res


def seg_to_beam(name: str, seg: ET.Element, parent: Part, prev_bm: Beam, zv):
    metadata = dict()

    index = seg.attrib["index"]
    sec = parent.sections.get_by_name(seg.attrib["section_ref"])
    material_ref = seg.attrib.get("material_ref", None)
    if material_ref is None:
        raise ValueError(f"Material not found for beam '{name}'. Please check your xml file")

    mat = parent.materials.get_by_name(material_ref)
    mdf = seg.attrib.get("mass_density_factor_ref", None)
    if mdf is not None:
        metadata["mass_density_factor_ref"] = mdf

    pos = {p.attrib["end"]: xyz_to_floats(p) for p in seg.findall(".//position")}
    if "reinforcement_ref" in seg.attrib.keys():
        metadata["reinforced"] = True

    if index != "1":
        name += f"_E{index}"
        if "cone" in seg.attrib["section_ref"].lower():
            sec = prev_bm.section

    n1 = parent.nodes.add(Node(pos_to_floats(pos["1"])))
    n2 = parent.nodes.add(Node(pos_to_floats(pos["2"])))

    try:
        bm = Beam(name, n1, n2, sec=sec, mat=mat, parent=parent, metadata=metadata, up=zv)
    except VectorNormalizeError:
        logger.warning(f"Beam '{name}' has coincident nodes. Will skip for now")
        return None

    return bm


def xyz_to_floats(p: ET.Element) -> Tuple[float]:
    return tuple(pos_to_floats([p.attrib["x"], p.attrib["y"], p.attrib["z"]]))


def pos_to_floats(pos):
    return [float(x) for x in pos]


def get_curve_orientation(root: ET.Element) -> tuple:
    def to_floats(v):
        return float(v.attrib["x"]), float(v.attrib["y"]), float(v.attrib["z"])

    zv = None
    yv = None
    xv = None
    lsys = root.find("./curve_orientation/customizable_curve_orientation/orientation/local_system")
    if lsys is not None:
        xvec = lsys.find("./xvector")
        yvec = lsys.find("./yvector")
        zvec = lsys.find("./zvector")
        xv = to_floats(xvec)
        yv = to_floats(yvec)
        zv = to_floats(zvec)
    else:
        lsys = root.find("./local_system")
        for vec in lsys.findall("./vector"):
            direction = vec.attrib.get("dir")
            if direction == "z":
                zv = to_floats(vec)
            elif direction == "y":
                yv = to_floats(vec)
            elif direction == "x":
                xv = to_floats(vec)

    if zv is None:
        raise ValueError("Z or Y vector must be set")

    return xv, yv, zv



from typing import TYPE_CHECKING, List

import numpy as np

from ada.fem import FemSet

if TYPE_CHECKING:
    from ada import FEM


def _read_families(fas_data):
    families = {}
    for _, node_set in fas_data.items():
        set_id = node_set.attrs["NUM"]  # unique set id
        n_subsets = node_set["GRO"].attrs["NBR"]  # number of subsets
        nom_dataset = node_set["GRO"]["NOM"][()]  # (n_subsets, 80) of int8
        name = [None] * n_subsets
        for i in range(n_subsets):
            name[i] = "".join([chr(x) for x in nom_dataset[i]]).strip().rstrip("\x00")
        families[set_id] = name
    return families


def _cell_tag_to_set(cell_data_array, cell_tags):
    """
    For a single element type convert tag data into set data

    :param cell_data_array:
    :param cell_tags:
    :return: Cell Sets dictionary
    """
    cell_sets = dict()
    shared_sets = []
    for tag_id, tag_names in cell_tags.items():
        if len(tag_names) > 1:
            for v in tag_names:
                res = np.where(cell_data_array == tag_id)[0]
                if len(res) > 0:
                    shared_sets.append((v, res))
        else:
            tag_name = tag_names[0]
            res = np.where(cell_data_array == tag_id)[0]
            if len(res) > 0:
                cell_sets[tag_name] = res

    for v, s in shared_sets:
        if v in cell_sets.keys():
            cell_sets[v] = np.concatenate([cell_sets[v], s])
        else:
            cell_sets[v] = s

    return cell_sets


def _point_tags_to_sets(tags, point_tags, fem: "FEM"):
    """

    :param tags:
    :param point_tags:
    :return: Point sets dictionary
    """
    point_sets = dict()
    shared_sets = []
    for key, val in point_tags.items():
        if len(val) > 1:
            for set_name in val:
                shared_sets.append((set_name, np.where(tags == key)[0]))
        else:
            point_sets[val[0]] = np.where(tags == key)[0]

    for set_name, s in shared_sets:
        point_sets[set_name] = np.concatenate([point_sets[set_name], s])

    nsets = [FemSet(pn, [fem.nodes.from_id(i + 1) for i in ps], "nset", parent=fem) for pn, ps in point_sets.items()]
    return nsets


def _element_set_dict_to_list_of_femset(element_sets, fem: "FEM") -> List[FemSet]:
    elsets = []
    for name, values in element_sets.items():
        elsets.append(FemSet(name, values, "elset", parent=fem))
    return elsets



from OCC.Extend.DataExchange import read_step_file
from OCC.Extend.TopologyUtils import TopologyExplorer


def iter_subshapes(shape):
    t = TopologyExplorer(shape)
    for solid in t.solids():
        yield solid
    for shell in t.shells():
        yield shell
    for face in t.faces():
        yield face


def make_conversion(stp_file):
    for x in iter_subshapes(read_step_file(stp_file, as_compound=False)):
        print(x)


if __name__ == "__main__":
    make_conversion("../../../files/step_files/Ventilator.stp")
